#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import time
from functools import partial
from copy import deepcopy
from pathlib import Path
import numpy as np
import tifffile
from scipy.optimize import (differential_evolution as deo,
                            dual_annealing as dao,
                            shgo, brute, fmin)
from scipy.ndimage import (zoom, shift, #rotate,
                           binary_erosion, fourier_shift,
                           gaussian_filter as gf)
from imutils import rotate

from skimage.registration import phase_cross_correlation
from pystackreg import StackReg

from .misc import msgit


def cost_norm(tgt, src, mask=1, gfk=3, type='tv', order='i', filt=True):
    """
    INPUTS:
        tgt: reference image
        src: target image
        type: optional
            'tv': total variance of the difference map between ref and tgt
            'de': difference energy between ref and tgt in L2 form
    """
    if not isinstance(mask, np.ndarray):
        mask = (src != 0)
        mask[:] = binary_erosion(mask, iterations=3)[:]
    pnts = mask.sum()

    if filt:
        if type == 'tv':
            if order == 'i':
                diff_img = gf(tgt, gfk) - gf(src, gfk)
                return (((np.abs(np.diff(diff_img, axis=0, prepend=1)) + np.abs(
                    np.diff(diff_img, axis=1, prepend=1)))) * mask).sum() / pnts
            elif order == 'ao':
                diff_img = (np.abs(
                    np.abs(np.diff(gf(tgt, gfk), axis=0, prepend=1)) - np.abs(np.diff(gf(src, gfk), axis=0, prepend=1))) + \
                            np.abs(np.abs(np.diff(gf(tgt, gfk), axis=1, prepend=1)) - np.abs(
                                np.diff(gf(src, gfk), axis=1, prepend=1))))
                return (diff_img * mask).sum() / pnts
            elif order == 'ai':
                diff_img = (np.abs(
                    np.abs(np.diff(gf(tgt, gfk), axis=0, prepend=1)) - np.abs(np.diff(gf(src, gfk), axis=1, prepend=1))) + \
                            np.abs(np.abs(np.diff(gf(tgt, gfk), axis=1, prepend=1)) - np.abs(
                                np.diff(gf(src, gfk), axis=0, prepend=1))))
                return (diff_img * mask).sum() / pnts
        elif type == 'de':
            return (((tgt - src) ** 2) * mask).sum() / pnts
    else:
        if type == 'tv':
            if order == 'i':
                diff_img = tgt - src
                return (((np.abs(np.diff(diff_img, axis=0, prepend=1)) + np.abs(
                    np.diff(diff_img, axis=1, prepend=1)))) * mask).sum() / pnts
            elif order == 'ao':
                diff_img = (np.abs(
                    np.abs(np.diff(tgt, axis=0, prepend=1)) - np.abs(np.diff(src, axis=0, prepend=1))) + \
                            np.abs(
                                np.abs(np.diff(tgt, axis=1, prepend=1)) - np.abs(np.diff(src, axis=1, prepend=1))))
                return (diff_img * mask).sum() / pnts
            elif order == 'ai':
                diff_img = (np.abs(
                    np.abs(np.diff(tgt, axis=0, prepend=1)) - np.abs(np.diff(src, axis=1, prepend=1))) + \
                            np.abs(
                                np.abs(np.diff(tgt, axis=1, prepend=1)) - np.abs(np.diff(src, axis=0, prepend=1))))
                return (diff_img * mask).sum() / pnts
        elif type == 'de':
            return (((tgt - src) ** 2) * mask).sum() / pnts


def cost_abs(tgt, src, mask=1, gfk=3, type='tv', order='i', filt=True):
    """
    INPUTS:
        tgt: reference image
        src: target image
        type: optional
            'tv': total variance of the difference map between ref and tgt
            'de': difference energy between ref and tgt in L2 form
    """
    if filt:
        if type == 'tv':
            if order == 'i':
                diff_img = gf(tgt, gfk) - gf(src, gfk)
                return (((np.abs(np.diff(diff_img, axis=0, prepend=1)) +
                          np.abs(np.diff(diff_img, axis=1, prepend=1)))) * mask).sum()
            elif order == 'ao':
                diff_img = (np.abs(
                    np.abs(np.diff(gf(tgt, gfk), axis=0, prepend=1)) - np.abs(np.diff(gf(src, gfk), axis=0, prepend=1))) + \
                            np.abs(np.abs(np.diff(gf(tgt, gfk), axis=1, prepend=1)) - np.abs(
                                np.diff(gf(src, gfk), axis=1, prepend=1))))
                return (diff_img * mask).sum()
            elif order == 'ai':
                diff_img = (np.abs(
                    np.abs(np.diff(gf(tgt, gfk), axis=0, prepend=1)) - np.abs(np.diff(gf(src, gfk), axis=1, prepend=1))) + \
                            np.abs(np.abs(np.diff(gf(tgt, gfk), axis=1, prepend=1)) - np.abs(
                                np.diff(gf(src, gfk), axis=0, prepend=1))))
                return (diff_img * mask).sum()
        elif type == 'de':
            return (((tgt - src) ** 2) * mask).sum()
    else:
        if type == 'tv':
            if order == 'i':
                diff_img = tgt - src
                return (((np.abs(np.diff(diff_img, axis=0, prepend=1)) +
                          np.abs(np.diff(diff_img, axis=1, prepend=1)))) * mask).sum()
            elif order == 'ao':
                diff_img = (np.abs(
                    np.abs(np.diff(tgt, axis=0, prepend=1)) - np.abs(np.diff(src, axis=0, prepend=1))) + \
                            np.abs(
                                np.abs(np.diff(tgt, axis=1, prepend=1)) - np.abs(np.diff(src, axis=1, prepend=1))))
                return (diff_img * mask).sum()
            elif order == 'ai':
                diff_img = (np.abs(
                    np.abs(np.diff(tgt, axis=0, prepend=1)) - np.abs(np.diff(src, axis=1, prepend=1))) + \
                            np.abs(
                                np.abs(np.diff(tgt, axis=1, prepend=1)) - np.abs(np.diff(src, axis=0, prepend=1))))
                return (diff_img * mask).sum()
        elif type == 'de':
            return (((tgt - src) ** 2) * mask).sum()


def tran(x, tgt, src, mask, gfk, roi=None, cft='tv', order='i'):
    """ tv of the difference map between ref and transformed tgt
    INPUTS:
       x: transformation parameters;
          x[0] and x[1]: x and y translations
       ref: reference image to be compared to
       tgt: the image to be transformed to be compared to ref
       gfk: gaussian_filter kernel
    RETURNS:
        TV: total variance of the difference map between ref and the
        transformed tgt
    """
    if roi is None:
        roi = np.s_[int((src.shape[0] - tgt.shape[0]) / 2):int((src.shape[0] - tgt.shape[0]) / 2) + tgt.shape[0],
              int((src.shape[1] - tgt.shape[1]) / 2):int((src.shape[1] - tgt.shape[1]) / 2) + tgt.shape[1]]
    return cost_norm(tgt, shift(src, x[:])[roi], mask=mask, gfk=gfk, type=cft, order=order)


def rota(x, tgt, src, mask, gfk, roi=None, cft='tv', order='i'):
    """ tv of the difference map between ref and transformed tgt
    INPUTS:
       x: transformation parameters;
          x[0]: rotaiton angle
       ref: reference image to be compared to
       tgt: the image to be transformed to be compared to ref
       gfk: gaussian_filter kernel
    RETURNS:
        TV: total variance of the difference map between ref and the
        transformed tgt
    """
    if roi is None:
        roi = np.s_[int((src.shape[0] - tgt.shape[0]) / 2):int((src.shape[0] - tgt.shape[0]) / 2) + tgt.shape[0],
              int((src.shape[1] - tgt.shape[1]) / 2):int((src.shape[1] - tgt.shape[1]) / 2) + tgt.shape[1]]
    # return cost_norm(tgt, rotate(src, x[0], order=1, reshape=False)[roi], mask=mask, gfk=gfk, type=cft, order=order)
    return cost_norm(tgt, rotate(src, x[0])[roi], mask=mask, gfk=gfk, type=cft, order=order)


def zoomu(x, ref, tgt, mask, gfk, roi=None, cft='tv', order='i'):
    """ tv of the difference map between ref and transformed tgt
    INPUTS:
       x: transformation parameters;
          x[0]: scaling factor
       ref: reference image to be compared to
       tgt: the image to be transformed to be compared to ref
       gfk: gaussian_filter kernel
    RETURNS:
        TV: total variance of the difference map between ref and the
        transformed tgt
    """
    tgt = zoom(tgt, x[0])
    if roi is None:
        roi = np.s_[int((tgt.shape[0] - ref.shape[0]) / 2):int((tgt.shape[0] - ref.shape[0]) / 2) + ref.shape[0],
              int((tgt.shape[1] - ref.shape[1]) / 2):int((tgt.shape[1] - ref.shape[1]) / 2) + ref.shape[1]]
    return cost_norm(ref, tgt[roi], gfk=gfk, type=cft, order=order)


def set_tran_quad(t):
    tm = np.eye(3)
    tm[0, 2] = t[0]
    tm[1, 2] = t[1]
    return tm


def set_rota_quad(r):
    rm = np.eye(3)
    rm[0, 0] = rm[1, 1] = np.cos(r)
    rm[0, 1] = -np.sin(r)
    rm[1, 0] = np.sin(r)
    return rm


def cal_ang_rad(c, s):
    if s >= 0:
        return np.arccos(c)
    elif s < 0:
        return 2 * np.pi - np.arccos(c)


def rad_2_deg(rad):
    return 180 * rad / np.pi


def deg_2_rad(deg):
    return deg * np.pi / 180


def rig_transform(src, param, dtype='mrtv'):
    """
    param: rigid-body transform paprameters
    dtype: string, optional; choose between 'pc', 'sr', 'mrtv'
    """
    if dtype == 'mrtv':
        return shift(rotate(src, rad_2_deg(cal_ang_rad(param[0, 0], param[1, 0]))), param[:2, 2])
    elif dtype == 'pc':
        return shift(rotate(src, -rad_2_deg(param[2])), [-param[1], -param[0]])
    elif dtype == 'sr':
        sr = StackReg(StackReg.RIGID_BODY)
        return sr.transform(src, param)


def mrtv_rigid_reg(tgt, src, pxl_conf, sub_conf, angr, itr=6, ps=None, mask=1, roi=None, gfk=1.5, popsize=27,
                   order='i'):
    s = deepcopy(src)
    m = np.eye(3)
    order = order

    tp, tpi, ss, mrtv_shift = mrtv_reg(pxl_conf, sub_conf, ps=ps, kernel=gfk, imgs=[tgt, s])
    m = np.matmul(set_tran_quad(mrtv_shift), m)
    print("0:\t", mrtv_shift, '\n')
    s[:] = shift(s, mrtv_shift)

    rlt = deo(rota, [list(angr)], args=(tgt, s, mask, gfk, roi, 'tv', order), workers=-1, popsize=popsize)
    m = np.matmul(set_rota_quad(np.pi * rlt.x[-1] / 180), m)
    print("0:\t", rlt.x, "\n")
    s[:] = rotate(s, rlt.x[-1])[:]

    for ii in range(itr):
        tp, tpi, ss, mrtv_shift = mrtv_reg(pxl_conf, sub_conf, ps=None, kernel=gfk, imgs=[tgt, s])
        m = np.matmul(set_tran_quad(mrtv_shift), m)
        print(f"{ii + 1}:\t", mrtv_shift, '\n')
        s[:] = shift(s, mrtv_shift)
        rlt = deo(rota, [(-10, 10)], args=(tgt, s, mask, gfk, roi, 'tv', order), workers=-1, popsize=popsize)
        m = np.matmul(set_rota_quad(np.pi * rlt.x[-1] / 180), m)
        print(f"{ii + 1}:\t", rlt.x, "\n")
        s[:] = rotate(s, rlt.x[-1])[:]
    return m, s


def mrtv_rigid_reg2(tgt, src, pxl_conf, sub_conf, angr, itr=6, ps=None, mask=1, roi=None, gfk=1.5, popsize=27,
                    order='i'):
    s = deepcopy(src)
    m = np.eye(3)
    order = order

    gm = (s != 0).astype(np.int8)
    tp, tpi, ss, mrtv_shift = mrtv_reg_v2(pxl_conf, sub_conf, ps=ps, gm=gm, gfk=gfk, imgs=[tgt, s])
    m = np.matmul(set_tran_quad(mrtv_shift), m)
    print("0:\t", mrtv_shift, '\n')
    s[:] = shift(s, np.array(mrtv_shift))

    rlt = deo(rota, [list(angr)], args=(tgt, s, mask, gfk, roi, 'tv', order), workers=-1, popsize=popsize)
    m = np.matmul(set_rota_quad(np.pi * rlt.x[-1] / 180), m)
    print("0:\t", rlt.x, "\n")
    s[:] = rotate(s, rlt.x[-1])[:]

    for ii in range(itr):
        gm = (s != 0).astype(np.int8)
        tp, tpi, ss, mrtv_shift = mrtv_reg_v2(pxl_conf, sub_conf, ps=None, gm=gm, gfk=gfk, imgs=[tgt, s])
        m = np.matmul(set_tran_quad(mrtv_shift), m)
        print(f"{ii + 1}:\t", mrtv_shift, '\n')
        s[:] = shift(s, np.array(mrtv_shift))

        rlt = deo(rota, [(-10, 10)], args=(tgt, s, mask, gfk, roi, 'tv', order), workers=-1, popsize=popsize)
        m = np.matmul(set_rota_quad(np.pi * rlt.x[-1] / 180), m)
        print(f"{ii + 1}:\t", rlt.x, "\n")
        s[:] = rotate(s, rlt.x[-1])[:]
    return m, s


def mrtv_rigid_reg4(tgt, src, pxl_conf, sub_conf, angr, itr=6, ps=None, mask=1, roi=None, gfk=1.5, popsize=27,
                    order='i', filt=True, norm=True):
    fpath = "/run/media/xiao/Data/data/tmp/gm_{}.tiff"
    s = deepcopy(src)
    m = np.eye(3)
    order = order

    rlt = deo(rota, [list(angr)], args=(tgt, s, mask, gfk, roi, 'tv', order), workers=-1, popsize=popsize)
    m = np.matmul(set_rota_quad(np.pi * rlt.x[-1] / 180), m)
    print("0:\t", rlt.x, "\n")
    s[:] = rotate(s, rlt.x[-1])[:]

    for ii in range(itr):
        gm = ((np.abs(s) > 1e-3)*(np.abs(tgt) > 1e-3)).astype(np.int8)
        tifffile.imsave(fpath.format(str(2 * ii).zfill(2)), gm)
        tp, tpi, ss, mrtv_shift = mrtv_reg_v4(pxl_conf, sub_conf, ps=None, gfk=gfk,
                                              gm=gm, imgs=[s, tgt], filt=filt, norm=norm)
        m = np.matmul(set_tran_quad(-np.array(mrtv_shift)), m)
        print(f"{ii + 1}:\t", mrtv_shift, '\n')
        s[:] = shift(s, -np.array(mrtv_shift))
        tifffile.imsave(Path(fpath.format(str(2 * ii + 1).zfill(2))), (np.abs(s) > 1e-3).astype(np.int8))

        rlt = deo(rota, [(-10, 10)], args=(tgt, s, mask, gfk, roi, 'tv', order), workers=-1, popsize=popsize)
        m = np.matmul(set_rota_quad(np.pi * rlt.x[-1] / 180), m)
        print(f"{ii + 1}:\t", rlt.x, "\n")
        s[:] = rotate(s, rlt.x[-1])[:]
    return m, s


def mrtv_rigid_reg5(tgt, src, pxl_conf, sub_conf, angr, itr=6, ps=None, mask=1, roi=None, gfk=1.5, popsize=27,
                    order='i'):
    fpath = "/run/media/xiao/Data/data/tmp/gm_{}.tiff"
    s = deepcopy(src)
    m = np.eye(3)
    order = order

    gm = (np.abs(s) > 1e-3).astype(np.int8)
    tp, tpi, ss, mrtv_shift = mrtv_reg_v4(pxl_conf, sub_conf, ps=ps, gm=gm, gfk=gfk, imgs=[s, tgt])
    m = np.matmul(set_tran_quad(-np.array(mrtv_shift)), m)
    print("0:\t", mrtv_shift, '\n')
    s[:] = shift(s, -np.array(mrtv_shift))

    gm = (np.abs(s) > 1e-3).astype(np.int8)
    rlt = deo(rota, [[-angr[1], -angr[0]]], args=(s, tgt, gm, gfk, roi, 'tv', order), workers=-1, popsize=popsize)
    m = np.matmul(set_rota_quad(-np.pi * rlt.x[-1] / 180), m)
    print("0:\t", rlt.x, "\n")
    s[:] = rotate(s, -rlt.x[-1])[:]

    for ii in range(itr):
        gm = (np.abs(s) > 1e-3).astype(np.int8)
        tifffile.imsave(Path(fpath.format(str(2*ii).zfill(2))), gm)
        tp, tpi, ss, mrtv_shift = mrtv_reg_v4(pxl_conf, sub_conf, ps=None, gm=gm, gfk=gfk, imgs=[s, tgt])
        m = np.matmul(set_tran_quad(-np.array(mrtv_shift)), m)
        print(f"{ii + 1}:\t", mrtv_shift, '\n')
        s[:] = shift(s, -np.array(mrtv_shift))

        gm = (np.abs(s) > 1e-3).astype(np.int8)
        tifffile.imsave(Path(fpath.format(str(2*ii+1).zfill(2))), gm)
        rlt = deo(rota, [(-10, 10)], args=(s, tgt, gm, gfk, roi, 'tv', order), workers=-1, popsize=popsize)
        m = np.matmul(set_rota_quad(-np.pi * rlt.x[-1] / 180), m)
        print(f"{ii + 1}:\t", rlt.x, "\n")
        s[:] = rotate(s, -rlt.x[-1])[:]
    return m, s


def tv_l1_pixel(fixed_img, img, mask, shift, filt=True, kernel=3):
    if filt:
        diff_img = gf(fixed_img, kernel) - \
            gf(np.roll(img, shift, axis=[0, 1]), kernel)
        return ((np.abs(np.diff(diff_img, axis=0, prepend=1))+np.abs(np.diff(diff_img, axis=1, prepend=1)))*np.roll(mask, shift, axis=[0, 1])).sum()
    else:
        diff_img = fixed_img - np.roll(img, shift, axis=[0, 1])
        return ((np.abs(np.diff(diff_img, axis=0, prepend=1))+np.abs(np.diff(diff_img, axis=1, prepend=1)))*np.roll(mask, shift, axis=[0, 1])).sum()


def tv_l1_pixel_v2(tgt, src, mask, gfk, shift, type='tv', order='l1'):
    """
    INPUTS:
        tgt: the reference image
        src: the image to be transformed to match tgt
    """
    s = np.roll(src, shift, axis=[0, 1])
    m = np.roll(mask, shift, axis=[0, 1])
    return cost_norm(tgt, s, mask=m, gfk=gfk, type='tv', order='l1')


def tv_l1_pixel_v3(tgt, src, mask, shift):
    """
    INPUTS:
        tgt: the reference image
        src: the image to be transformed to match tgt
    """
    diff_img = tgt-np.roll(src, shift, axis=[0, 1])
    return ((np.abs(np.diff(diff_img, axis=0, prepend=1))+np.abs(np.diff(diff_img, axis=1, prepend=1)))*mask).sum()


def tv_l1_pixel_v4(fixed_img, img, mask, shift, filt=True, kernel=3, norm=True):
    if norm:
        return cost_norm(fixed_img, np.roll(img, shift, axis=[0, 1]), mask=mask, gfk=kernel, type='tv', order='i', filt=filt)
    else:
        return cost_abs(fixed_img, np.roll(img, shift, axis=[0, 1]), mask=mask, gfk=kernel, type='tv', order='i', filt=filt)


def tv_l2_pixel(fixed_img, img, mask, shift, norm=True):
    if norm:
        diff_img = fixed_img - np.roll(img, shift, axis=[0, 1])
        diff_img /= np.sqrt((diff_img**2).sum())
    else:
        diff_img = fixed_img - np.roll(img, shift, axis=[0, 1])
    return (((np.diff(diff_img, axis=0, prepend=1))**2 +
             (np.diff(diff_img, axis=1, prepend=1))**2)*mask).sum()


def tv_l1_subpixel(fixed_img, img, mask, shift, filt=True):
    if filt:
        diff_img = (gf(fixed_img, 3) -
                    np.real(np.fft.ifftn(fourier_shift(np.fft.fftn(gf(img, 3)), shift))))
        return ((np.abs(np.diff(diff_img, axis=0, prepend=1)) +
                 np.abs(np.diff(diff_img, axis=1, prepend=1)))*mask).sum()
    else:
        diff_img = fixed_img - \
            np.real(np.fft.ifftn(fourier_shift(np.fft.fftn(img), shift)))
        return ((np.abs(np.diff(diff_img, axis=0, prepend=1)) +
                 np.abs(np.diff(diff_img, axis=1, prepend=1)))*mask).sum()


def tv_l1_subpixel_fit(x0, x1, tvn):
    coef = np.polyfit(x0, tvn[0], 2)
    tvx = x1[np.argmin(np.polyval(coef, x1))]
    coef = np.polyfit(x0, tvn[1], 2)
    tvy = x1[np.argmin(np.polyval(coef, x1))]
    return [tvx, tvy]


def tv_l1_subpixel_ana(x3, x21, tv3):
    a = (tv3[0][0]+tv3[0][2]-2*tv3[0][1])/2
    b = (tv3[0][2]-tv3[0][0])/2
    c = tv3[0][1]
    tvy = x21[np.argmin(a*x21**2 + b*x21 + c)]
    a = (tv3[1][0]+tv3[1][2]-2*tv3[1][1])/2
    b = (tv3[1][2]-tv3[1][0])/2
    c = tv3[1][1]
    tvx = x21[np.argmin(a*x21**2 + b*x21 + c)]
    return [tvy, tvx]


def tv_l2_subpixel(fixed_img, img, mask, shift, norm=True):
    if norm:
        diff_img = fixed_img - \
            np.real(np.fft.ifftn(fourier_shift(np.fft.fftn(img), shift)))
        diff_img /= np.sqrt((diff_img**2).sum())
    else:
        diff_img = fixed_img - \
            np.real(np.fft.ifftn(fourier_shift(np.fft.fftn(img), shift)))
    return (((np.diff(diff_img, axis=0, prepend=1))**2 +
             (np.diff(diff_img, axis=1, prepend=1))**2)*mask).sum()


def mrtv_reg1(fixed_img, img, levs=4, wz=10, sp_wz=20, sp_step=0.2):
    wz = int(wz)
    sch_config = {}
    sch_config[levs-1] = {'wz': sp_wz, 'step': sp_step}
    tv1_pxl = {}
    tv1_pxl_id = np.zeros(levs, dtype=np.int16)
    shift = np.zeros([levs, 2], dtype=np.float32)
    rlt = []

    for ii in range(levs-1):
        sch_config[levs-2-ii] = {'wz': wz, 'step': int(2**ii)}

    border_wz = sch_config[0]['wz']*sch_config[0]['step']
    if ((np.array(fixed_img.shape) - border_wz) <= 0).any():
        return -1

    mask = np.zeros(fixed_img.shape)
    mask[int(border_wz/2):-int(border_wz/2),
         int(border_wz/2):-int(border_wz/2)] = 1

    for ii in range(levs-1):
        w = sch_config[ii]['wz']
        step = sch_config[ii]['step']
        with mp.get_context('spawn').Pool(N_CPU) as pool:
            rlt = pool.map(partial(tv_l1_pixel, fixed_img, img, mask),
                           [[int(step*(jj//w-int(w/2))+shift[:ii, 0].sum()),
                             int(step*(jj % w-int(w/2))+shift[:ii, 1].sum())]
                            for jj in np.int32(np.arange(w**2))])
        pool.close()
        pool.join()

        tem = np.ndarray([w, w], dtype=np.float32)
        for kk in range(w**2):
            tem[kk//w, kk % w] = rlt[kk]
        del(rlt)
        gc.collect()

        tv1_pxl[ii] = np.array(tem)
        tv1_pxl_id[ii] = tv1_pxl[ii].argmin()
        shift[ii, 0] = step*(tv1_pxl_id[ii]//w-int(w/2))
        shift[ii, 1] = step*(tv1_pxl_id[ii] % w-int(w/2))

    w = sch_config[levs-1]['wz']
    step = sch_config[levs-1]['step']
    with mp.get_context('spawn').Pool(N_CPU) as pool:
        rlt = pool.map(partial(tv_l1_subpixel, fixed_img, img, mask),
                       [[step*(jj//w-int(w/2))+shift[:(levs-1), 0].sum(),
                         step*(jj % w-int(w/2))+shift[:(levs-1), 1].sum()]
                        for jj in np.int32(np.arange(w**2))])
    pool.close()
    pool.join()

    tem = np.ndarray([w, w], dtype=np.float32)
    for kk in range(w**2):
        tem[kk//w, kk % w] = rlt[kk]
    del(rlt)
    gc.collect()

    tv1_pxl[levs-1] = np.array(tem)
    tv1_pxl_id[levs-1] = tv1_pxl[levs-1].argmin()
    shift[levs-1, 0] = step*(tv1_pxl_id[levs-1]//w-int(w/2))
    shift[levs-1, 1] = step*(tv1_pxl_id[levs-1] % w-int(w/2))

    print(time.asctime())
    return tv1_pxl, tv1_pxl_id, shift, shift.sum(axis=0)


def mrtv_reg2(fixed_img, img, levs=4, wz=10, sp_wz=20, sp_step=0.2, ps=None):
    if ps is not None:
        img[:] = np.real(np.fft.ifftn(fourier_shift(np.fft.fftn(img), ps)))[:]
    wz = int(wz)
    sch_config = {}
    sch_config[levs-1] = {'wz': sp_wz, 'step': sp_step}
    tv1_pxl = {}
    tv1_pxl_id = np.zeros(levs, dtype=np.int16)
    shift = np.zeros([levs, 2], dtype=np.float32)
    rlt = []

    for ii in range(levs-1):
        sch_config[levs-2-ii] = {'wz': 6, 'step': 0.5**ii}
    sch_config[0] = {'wz': wz, 'step': 0.5**(levs-2)}

    if ((np.array(fixed_img.shape)*0.5**(levs-2) - wz) <= 0).any():
        return -1

    for ii in range(levs-1):
        w = sch_config[ii]['wz']
        step = sch_config[ii]['step']
        f = zoom(fixed_img, step)
        m = zoom(img, step)
        mk = np.zeros(m.shape, dtype=np.int8)
        mk[int(wz*2**(ii-1)):-int(wz*2**(ii-1)),
           int(wz*2**(ii-1)):-int(wz*2**(ii-1))] = 1

        s = np.array([0, 0])
        for kk in range(ii):
            s = s + shift[kk]*2**(ii-kk)
        with mp.get_context('spawn').Pool(N_CPU) as pool:
            rlt = pool.map(partial(tv_l1_pixel, f, m, mk),
                           [[int((jj//w-int(w/2))+s[0]),
                             int((jj % w-int(w/2))+s[1])]
                            for jj in np.int32(np.arange(w**2))])
        pool.close()
        pool.join()

        tem = np.ndarray([w, w], dtype=np.float32)
        for kk in range(w**2):
            tem[kk//w, kk % w] = rlt[kk]
        del(rlt)
        gc.collect()

        tv1_pxl[ii] = np.array(tem)
        tv1_pxl_id[ii] = tv1_pxl[ii].argmin()
        shift[ii, 0] = (tv1_pxl_id[ii]//w-int(w/2))
        shift[ii, 1] = (tv1_pxl_id[ii] % w-int(w/2))

    mk = np.zeros(fixed_img.shape)
    mk[int(wz*2**(levs-2)):-int(wz*2**(levs-2)),
       int(wz*2**(levs-2)):-int(wz*2**(levs-2))] = 1
    w = sch_config[levs-1]['wz']
    step = sch_config[levs-1]['step']
    s = s + shift[levs-2]
    with mp.get_context('spawn').Pool(N_CPU) as pool:
        rlt = pool.map(partial(tv_l1_subpixel, fixed_img, img, mk),
                       [[step*(jj//w-int(w/2))+s[0],
                         step*(jj % w-int(w/2))+s[1]]
                        for jj in np.int32(np.arange(w**2))])
    pool.close()
    pool.join()

    tem = np.ndarray([w, w], dtype=np.float32)
    for kk in range(w**2):
        tem[kk//w, kk % w] = rlt[kk]
    del(rlt)
    gc.collect()

    tv1_pxl[levs-1] = np.array(tem)
    tv1_pxl_id[levs-1] = tv1_pxl[levs-1].argmin()
    shift[levs-1, 0] = step*(tv1_pxl_id[levs-1]//w-int(w/2))
    shift[levs-1, 1] = step*(tv1_pxl_id[levs-1] % w-int(w/2))

    print(time.asctime())
    return tv1_pxl, tv1_pxl_id, shift, s+shift[levs-1]


@msgit(wd=100, fill='-')
def mrtv_reg3(levs=7, wz=10, sp_wz=8, sp_step=0.5, ps=None, imgs=None):
    """
    single process version of mrtv_reg2. Parallelization is implemented
    on the upper level on which the alignments of pairs of images are
    parallelized
    """
    if ps is not None:
        imgs[1] = np.real(np.fft.ifftn(
            fourier_shift(np.fft.fftn(imgs[1]), ps)))

    sch_config = {}
    sch_config[levs-1] = {'wz': sp_wz, 'step': sp_step}
    tv1_pxl = {}
    tv1_pxl_id = np.zeros(levs, dtype=np.int16)
    shift = np.zeros([levs, 2], dtype=np.float32)

    for ii in range(levs-1):
        sch_config[levs-2-ii] = {'wz': 6, 'step': 0.5**ii}
    sch_config[0] = {'wz': wz, 'step': 0.5**(levs-2)}

    if ((np.array(imgs[0].shape)*0.5**(levs-2) - wz) <= 0).any():
        return -1

    for ii in range(levs-1):
        w = sch_config[ii]['wz']
        step = sch_config[ii]['step']
        f = zoom(imgs[0], step)
        m = zoom(imgs[1], step)
        mk = np.zeros(m.shape, dtype=np.int8)
        mk[int(wz*2**(ii-1)):-int(wz*2**(ii-1)),
           int(wz*2**(ii-1)):-int(wz*2**(ii-1))] = 1

        s = np.array([0, 0], dtype=np.int32)
        for jj in range(ii):
            s = np.int_(s + shift[jj]*2**(ii-jj))

        tem = np.ndarray([w, w], dtype=np.float32)
        for jj in range(w):
            for kk in range(w):
                tem[jj, kk] = tv_l1_pixel(f, m, mk,
                                          [jj-int(w/2)+s[0], kk-int(w/2)+s[1]])

        tv1_pxl[ii] = np.array(tem)
        tv1_pxl_id[ii] = tv1_pxl[ii].argmin()
        shift[ii, 0] = (tv1_pxl_id[ii]//w-int(w/2))
        shift[ii, 1] = (tv1_pxl_id[ii] % w-int(w/2))

    mk = np.zeros(imgs[0].shape)
    mk[int(wz*2**(levs-2)):-int(wz*2**(levs-2)),
       int(wz*2**(levs-2)):-int(wz*2**(levs-2))] = 1
    w = sch_config[levs-1]['wz']
    step = sch_config[levs-1]['step']
    s = s + shift[levs-2]

    tem = np.ndarray([w, w], dtype=np.float32)
    for jj in range(w):
        for kk in range(w):
            tem[jj, kk] = tv_l1_subpixel(imgs[0], imgs[1], mk,
                                         [jj-int(w/2)+s[0], kk-int(w/2)+s[1]])

    tv1_pxl[levs-1] = np.array(tem)
    tv1_pxl_id[levs-1] = tv1_pxl[levs-1].argmin()
    shift[levs-1, 0] = step*(tv1_pxl_id[levs-1]//w-int(w/2))
    shift[levs-1, 1] = step*(tv1_pxl_id[levs-1] % w-int(w/2))

    print(time.asctime())
    return tv1_pxl, tv1_pxl_id, shift, s+shift[levs-1]


@msgit(wd=100, fill='-')
def mrtv_reg4(levs=7, wz=10, sp_wz=8, sp_step=0.5, ps=None, imgs=None):
    """
    Single process version of mrtv_reg2. Parallelization is implemented
    on the upper level on which the alignments of pairs of images are
    parallelized

    INPUTS:
        levs: int
    """
    if ps is not None:
        imgs[1] = np.real(np.fft.ifftn(
            fourier_shift(np.fft.fftn(imgs[1]), ps)))

    sch_config = {}
    sch_config[levs-1] = {'wz': sp_wz, 'step': sp_step}
    tv1_pxl = {}
    tv1_pxl_id = np.zeros(levs, dtype=np.int16)
    shift = np.zeros([levs, 2], dtype=np.float32)

    for ii in range(levs-1):
        sch_config[levs-2-ii] = {'wz': wz, 'step': 0.5**ii}
    sch_config[0] = {'wz': wz, 'step': 0.5**(levs-2)}

    if ((np.array(imgs[0].shape)*0.5**(levs-2) - wz) <= 0).any():
        return -1

    for ii in range(levs-1):
        w = sch_config[ii]['wz']
        step = sch_config[ii]['step']
        f = zoom(imgs[0], step)
        m = zoom(imgs[1], step)
        mk = np.zeros(m.shape, dtype=np.int8)
        mk[int(wz*2**(ii-1)):-int(wz*2**(ii-1)),
           int(wz*2**(ii-1)):-int(wz*2**(ii-1))] = 1

        s = np.array([0, 0], dtype=np.int32)
        for jj in range(ii):
            s = np.int_(s + shift[jj]*2**(ii-jj))

        tem = np.ndarray([w, w], dtype=np.float32)
        for jj in range(w):
            for kk in range(w):
                tem[jj, kk] = tv_l1_pixel(f, m, mk,
                                          [jj-int(w/2)+s[0], kk-int(w/2)+s[1]])

        tv1_pxl[ii] = np.array(tem)
        tv1_pxl_id[ii] = tv1_pxl[ii].argmin()
        shift[ii, 0] = (tv1_pxl_id[ii]//w-int(w/2))
        shift[ii, 1] = (tv1_pxl_id[ii] % w-int(w/2))

    s = s + shift[levs-2]

    if sp_wz > (int(wz/2)-1):
        sp_wz = (int(wz/2)-1)

    idx = np.int_(shift[levs-2] + int(w/2))
    print('shift:', shift)
    print('idx:', idx)
    shift[levs-1] = tv_l1_subpixel_fit(np.linspace(-sp_wz, sp_wz, 2*sp_wz+1),
                                       np.linspace(-sp_wz*10, sp_wz *
                                                   10, 2*sp_wz*10+1)/10,
                                       [tv1_pxl[levs-2].reshape([wz, wz])[(idx[0]-sp_wz):(idx[0]+sp_wz+1), idx[1]],
                                        tv1_pxl[levs-2].reshape([wz, wz])[idx[0], (idx[1]-sp_wz):(idx[1]+sp_wz+1)]])

    print(time.asctime())
    return tv1_pxl, tv1_pxl_id, shift, s+shift[levs-1]


@msgit(wd=100, fill='-')
def mrtv_reg5(levs=7, wz=10, sp_wz=1, sp_step=0.5, lsw=10, ps=None, imgs=None):
    """
    single process version of mrtv_reg2. Parallelization is implemented
    on the upper level on which the alignments of pairs of images are
    parallelized
    """
    if ps is not None:
        imgs[1] = np.real(np.fft.ifftn(
            fourier_shift(np.fft.fftn(imgs[1]), ps)))

    sch_config = {}
    sch_config[levs-1] = {'wz': sp_wz, 'step': sp_step}
    tv1_pxl = {}
    tv1_pxl_id = np.zeros(levs, dtype=np.int16)
    shift = np.zeros([levs, 2], dtype=np.float32)

    for ii in range(levs-1):
        sch_config[levs-2-ii] = {'wz': wz, 'step': 0.5**ii}
    sch_config[0] = {'wz': wz, 'step': 0.5**(levs-2)}

    if ((np.array(imgs[0].shape)*0.5**(levs-2) - wz) <= 0).any():
        return -1

    for ii in range(levs-1):
        w = sch_config[ii]['wz']
        step = sch_config[ii]['step']
        f = zoom(imgs[0], step)
        m = zoom(imgs[1], step)
        mk = np.zeros(m.shape, dtype=np.int8)
        mk[int(wz*2**(ii-1)):-int(wz*2**(ii-1)),
           int(wz*2**(ii-1)):-int(wz*2**(ii-1))] = 1

        s = np.array([0, 0], dtype=np.int32)
        for jj in range(ii):
            s = np.int_(s + shift[jj]*2**(ii-jj))

        if ii == levs-2:
            tv_lv = []
            for kk in range(2*lsw+1):
                tv_lv.append(tv_l1_pixel(f, m, 1, [kk-lsw+s[0], s[1]]))
            s[0] += (np.array(tv_lv).argmin() - lsw)

            tv_lh = []
            for kk in range(2*lsw+1):
                tv_lh.append(tv_l1_pixel(f, m, 1, [s[0], kk-lsw+s[1]]))
            s[1] += (np.array(tv_lh).argmin() - lsw)

        tem = np.ndarray([w, w], dtype=np.float32)
        for jj in range(w):
            for kk in range(w):
                tem[jj, kk] = tv_l1_pixel(f, m, mk,
                                          [jj-int(w/2)+s[0], kk-int(w/2)+s[1]])

        tv1_pxl[ii] = np.array(tem)
        tv1_pxl_id[ii] = tv1_pxl[ii].argmin()
        shift[ii, 0] = (tv1_pxl_id[ii]//w-int(w/2))
        shift[ii, 1] = (tv1_pxl_id[ii] % w-int(w/2))

    print('s before:', s)
    s = s + shift[levs-2]

    if sp_wz > (int(wz/2)-1):
        sp_wz = (int(wz/2)-1)

    idx = np.int_(shift[levs-2] + int(w/2))
    print('s after:', s)
    print('shift:', shift)
    print('idx:', idx)

    us = int(1/sp_step)
    shift[levs-1] = tv_l1_subpixel_ana(np.linspace(-1, 1, 3),
                                       np.linspace(-us, us, 2*us+1)/us,
                                       [tv1_pxl[levs-2].reshape([wz, wz])[(idx[0]-sp_wz):(idx[0]+sp_wz+1), idx[1]],
                                        tv1_pxl[levs-2].reshape([wz, wz])[idx[0], (idx[1]-sp_wz):(idx[1]+sp_wz+1)]])
    return tv1_pxl, tv1_pxl_id, shift, s+shift[levs-1]


def mrtv_reg(pxl_conf, sub_conf, ps=None, kernel=3, imgs=None):
    """
    Provide a unified interace for using multi-resolution TV-based registration
    algorithm with different configuration options.

    Inputs:
    ______
        pxl_conf: dict; multi-resolution TV minimization configuration at
                  resolution higher than a single pixel; it includes items:
                  'type': 'area', 'line', 'al', 'la'
                  'levs': int
                  'wz': int
                  'lsw': int
        sub_conf: dict; sub-pixel TV minimization configuration; it includes
                  items:
                  'use': boolean; if conduct sub-pixel registration on the end
                  'type': ['ana', 'fit']; which sub-pixel routine to use
                  'sp_wz': int; for 'fit' option; the number of TV points to be
                           used in fitting
                  'sp_us': int; for 'ana' option; up-sampling factor
        ps: optonal; None or a 2-element tuple
        imgs: the pair of the images to be registered

    Outputs:
    _______
        tv_pxl: dict; TV at each search point under levels as the keywords
        tv_pxl_id: the id of the minimum TV in the flatted tv_pxl for each level
        shift: ndarray; shift at each level of resolution
        tot_shift: 2-element tuple; overall displacement between the pair of images
    """
    """
    single process version of mrtv_reg2. Parallelization is implemented
    on the upper level on which the alignments of pairs of images are
    parallelized
    """
    if ps is not None:
        imgs[1] = np.real(np.fft.ifftn(
            fourier_shift(np.fft.fftn(imgs[1]), ps)))

    levs = pxl_conf['levs']
    w = pxl_conf['wz']
    step = {}
    tv_pxl = {}
    tv_pxl_id = np.zeros(levs, dtype=np.int16)
    shift = np.zeros([levs+1, 2], dtype=np.float32)

    for ii in range(levs):
        step[levs-1-ii] = 0.5**ii

    if ((np.array(imgs[0].shape)*0.5**(levs-2) - w) <= 0).any():
        return -1

    if pxl_conf['type'] == 'area':
        for ii in range(levs):
            s = step[ii]
            f = zoom(imgs[0], s)
            m = zoom(imgs[1], s)
            mk = np.zeros(m.shape, dtype=np.int8)
            mk[int(w*2**(ii-1)):-int(w*2**(ii-1)),
               int(w*2**(ii-1)):-int(w*2**(ii-1))] = 1
            mk *= binary_erosion((m!=0)).astype(np.int8)

            sh = np.array([0, 0])
            for jj in range(ii):
                sh = np.int_(sh + shift[jj]*2**(ii-jj))

            tem = np.ndarray([w, w], dtype=np.float32)
            for jj in range(w):
                for kk in range(w):
                    tem[jj, kk] = tv_l1_pixel(f, m, mk,
                                              [jj-int(w/2)+sh[0],
                                               kk-int(w/2)+sh[1]],
                                              kernel=kernel)

            tv_pxl[ii] = np.array(tem)
            tv_pxl_id[ii] = tv_pxl[ii].argmin()
            shift[ii, 0] = (tv_pxl_id[ii]//w-int(w/2))
            shift[ii, 1] = (tv_pxl_id[ii] % w-int(w/2))
            if ii == levs - 1:
                idx = np.int_(shift[levs-1] + int(w/2))
                while idx[0] == 0 or idx[0] == w-1 or idx[1] == 0 or idx[1] == w-1:
                    if idx[0] == 0:
                        sh[0] -= int(w/2)
                    elif idx[0] == w-1:
                        sh[0] += int(w/2)
                    if idx[1] == 0:
                        sh[1] -= int(w/2)
                    elif idx[1] == w-1:
                        sh[1] += int(w/2)
                    for jj in range(w):
                        for kk in range(w):
                            tem[jj, kk] = tv_l1_pixel(f, m, mk,
                                                      [jj-int(w/2)+sh[0],
                                                       kk-int(w/2)+sh[1]],
                                                      kernel=kernel)

                    tv_pxl[ii] = np.array(tem)
                    tv_pxl_id[ii] = tv_pxl[ii].argmin()
                    shift[ii, 0] = (tv_pxl_id[ii]//w-int(w/2))
                    shift[ii, 1] = (tv_pxl_id[ii] % w-int(w/2))
                    idx = np.int_(shift[levs-1] + int(w/2))
        sh = sh + shift[levs-1]
    elif pxl_conf['type'] == 'line':
        levs = 1
        lsw = pxl_conf['lsw']
        shift = np.zeros([2, 2])
        tv_pxl_id = np.zeros(2, dtype=np.int16)

        tv_v = []
        mk = np.zeros(imgs[0].shape, dtype=np.int8)
        mk[int(lsw/2):-int(lsw/2), :] = 1
        for ii in range(lsw):
            tv_v.append(tv_l1_pixel(imgs[0], imgs[1], mk,
                                    [ii-int(lsw/2), 0],
                                    kernel=kernel))
        shift[0, 0] = (np.array(tv_v).argmin() - int(lsw/2))
        tv_h = []
        mk = np.zeros(imgs[0].shape, dtype=np.int8)
        mk[:, int(lsw/2):-int(lsw/2)] = 1
        for ii in range(lsw):
            tv_h.append(tv_l1_pixel(imgs[0], imgs[1], mk,
                                    [0, ii-int(lsw/2)],
                                    kernel=kernel))
        shift[0, 1] = (np.array(tv_h).argmin() - int(lsw/2))
        tv_pxl[0] = np.stack([np.array(tv_v), np.array(tv_h)], axis=0)
        tv_pxl_id[0] = 0
        sh = shift
    elif pxl_conf['type'] == 'al':
        lsw = pxl_conf['lsw']
        for ii in range(levs):
            s = step[ii]
            f = zoom(imgs[0], s)
            m = zoom(imgs[1], s)
            mk = np.zeros(m.shape, dtype=np.int8)
            mk[int(w*2**(ii-1)):-int(w*2**(ii-1)),
               int(w*2**(ii-1)):-int(w*2**(ii-1))] = 1

            sh = np.array([0, 0])
            for jj in range(ii):
                sh = np.int_(sh + shift[jj]*2**(ii-jj))

            if ii == levs-1:
                tv_v = []
                mk = np.zeros(imgs[0].shape, dtype=np.int8)
                mk[int(lsw/2):-int(lsw/2), :] = 1
                for kk in range(lsw):
                    tv_v.append(tv_l1_pixel(f, m, mk,
                                            [kk-int(lsw/2)+sh[0], sh[1]],
                                            kernel=kernel))
                sh[0] += (np.array(tv_v).argmin() - int(lsw/2))

                tv_h = []
                mk = np.zeros(imgs[0].shape, dtype=np.int8)
                mk[:, int(lsw/2):-int(lsw/2)] = 1
                for kk in range(lsw):
                    tv_h.append(tv_l1_pixel(f, m, mk,
                                            [sh[0], kk-int(lsw/2)+sh[1]],
                                            kernel=kernel))
                sh[1] += (np.array(tv_h).argmin() - int(lsw/2))

            tem = np.ndarray([w, w], dtype=np.float32)
            for jj in range(w):
                for kk in range(w):
                    tem[jj, kk] = tv_l1_pixel(f, m, mk,
                                              [jj-int(w/2)+sh[0],
                                               kk-int(w/2)+sh[1]],
                                              kernel=kernel)

            tv_pxl[ii] = np.array(tem)
            tv_pxl_id[ii] = tv_pxl[ii].argmin()
            shift[ii, 0] = (tv_pxl_id[ii]//w-int(w/2))
            shift[ii, 1] = (tv_pxl_id[ii] % w-int(w/2))

            if ii == levs - 1:
                idx = np.int_(shift[levs-1] + int(w/2))
                while idx[0] == 0 or idx[0] == w-1 or idx[1] == 0 or idx[1] == w-1:
                    if idx[0] == 0:
                        sh[0] -= int(w/2)
                    elif idx[0] == w-1:
                        sh[0] += int(w/2)
                    if idx[1] == 0:
                        sh[1] -= int(w/2)
                    elif idx[1] == w-1:
                        sh[1] += int(w/2)
                    for jj in range(w):
                        for kk in range(w):
                            tem[jj, kk] = tv_l1_pixel(f, m, mk,
                                                      [jj-int(w/2)+sh[0],
                                                       kk-int(w/2)+sh[1]],
                                                      kernel=kernel)

                    tv_pxl[ii] = np.array(tem)
                    tv_pxl_id[ii] = tv_pxl[ii].argmin()
                    shift[ii, 0] = (tv_pxl_id[ii]//w-int(w/2))
                    shift[ii, 1] = (tv_pxl_id[ii] % w-int(w/2))
                    idx = np.int_(shift[levs-1] + int(w/2))
        sh = sh + shift[levs-1]
    elif pxl_conf['type'] == 'la':
        # TODO
        pass

    if sub_conf['use']:
        idx = np.int_(shift[levs-1] + int(w/2))
        if sub_conf['type'] == 'fit':
            sw = int(sub_conf['sp_wz'])
            shift[levs] = tv_l1_subpixel_fit(np.linspace(-sw, sw, 2*sw+1),
                                             np.linspace(-10*sw,
                                                         10*sw, 20*sw+1),
                                             [tv_pxl[levs-1].reshape(w, w)
                                              [(idx[0]-sw):(idx[0]+sw+1), idx[1]],
                                              tv_pxl[levs-1].reshape(w, w)
                                              [idx[0], (idx[1]-sw):(idx[1]+sw+1)]])
        elif sub_conf['type'] == 'ana':
            us = int(sub_conf['sp_us'])
            shift[levs] = tv_l1_subpixel_ana(np.linspace(-1, 1, 3),
                                             np.linspace(-us, us, 2*us+1)/us,
                                             [tv_pxl[levs-1].reshape(w, w)
                                              [(idx[0]-1):(idx[0]+2), idx[1]],
                                              tv_pxl[levs-1].reshape(w, w)
                                              [idx[0], (idx[1]-1):(idx[1]+2)]])
    else:
        shift[levs] = [0, 0]
    return tv_pxl, tv_pxl_id, shift, sh+shift[levs]


def mrtv_reg_v2(pxl_conf, sub_conf, ps=None, gfk=3, gm=1, imgs=None):
    """
    Provide a unified interace for using multi-resolution TV-based registration
    algorithm with different configuration options. Using normalized TVDM on pixels

    Inputs:
    ______
        pxl_conf: dict; multi-resolution TV minimization configuration at
                  resolution higher than a single pixel; it includes items:
                  'type': 'area', 'line', 'al', 'la'
                  'levs': int
                  'wz': int
                  'lsw': int
        sub_conf: dict; sub-pixel TV minimization configuration; it includes
                  items:
                  'use': boolean; if conduct sub-pixel registration on the end
                  'type': ['ana', 'fit']; which sub-pixel routine to use
                  'sp_wz': int; for 'fit' option; the number of TV points to be
                           used in fitting
                  'sp_us': int; for 'ana' option; up-sampling factor
        ps: optonal; None or a 2-element tuple
        gm: optional; global mask
        imgs: the pair of the images to be registered

    Outputs:
    _______
        tv_pxl: dict; TV at each search point under levels as the keywords
        tv_pxl_id: the id of the minimum TV in the flatted tv_pxl for each level
        shift: ndarray; shift at each level of resolution
        tot_shift: 2-element tuple; overall displacement between the pair of images
    """
    """
    single process version of mrtv_reg2. Parallelization is implemented
    on the upper level on which the alignments of pairs of images are
    parallelized
    """
    if ps is not None:
        imgs[1] = np.real(np.fft.ifftn(
            fourier_shift(np.fft.fftn(imgs[1]), ps)))

    levs = pxl_conf['levs']
    w = pxl_conf['wz']
    step = {}
    tv_pxl = {}
    tv_pxl_id = np.zeros(levs, dtype=np.int16)
    shift = np.zeros([levs+1, 2], dtype=np.float32)

    for ii in range(levs):
        step[levs-1-ii] = 0.5**ii

    if ((np.array(imgs[0].shape)*0.5**(levs-2) - w) <= 0).any():
        return -1

    if pxl_conf['type'] == 'area':
        for ii in range(levs):
            s = step[ii]
            f = zoom(imgs[0], s)
            m = zoom(imgs[1], s)
            gmz = zoom(gm, s)
            mk = np.zeros(m.shape, dtype=np.int8)
            mk[int(w*2**(ii-1)):-int(w*2**(ii-1)),
               int(w*2**(ii-1)):-int(w*2**(ii-1))] = 1
            mk *= gmz

            sh = np.array([0, 0])
            for jj in range(ii):
                sh = np.int_(sh + shift[jj]*2**(ii-jj))

            tem = np.ndarray([w, w], dtype=np.float32)
            for jj in range(w):
                for kk in range(w):
                    tem[jj, kk] = tv_l1_pixel_v2(f, m, mk, gfk, [jj-int(w/2)+sh[0], kk-int(w/2)+sh[1]])

            tv_pxl[ii] = np.array(tem)
            tv_pxl_id[ii] = tv_pxl[ii].argmin()
            shift[ii, 0] = (tv_pxl_id[ii]//w-int(w/2))
            shift[ii, 1] = (tv_pxl_id[ii] % w-int(w/2))
            if ii == levs - 1:
                idx = np.int_(shift[levs-1] + int(w/2))
                while idx[0] == 0 or idx[0] == w-1 or idx[1] == 0 or idx[1] == w-1:
                    if idx[0] == 0:
                        sh[0] -= int(w/2)
                    elif idx[0] == w-1:
                        sh[0] += int(w/2)
                    if idx[1] == 0:
                        sh[1] -= int(w/2)
                    elif idx[1] == w-1:
                        sh[1] += int(w/2)
                    for jj in range(w):
                        for kk in range(w):
                            tem[jj, kk] = tv_l1_pixel_v2(f, m, mk, gfk, [jj-int(w/2)+sh[0], kk-int(w/2)+sh[1]])

                    tv_pxl[ii] = np.array(tem)
                    tv_pxl_id[ii] = tv_pxl[ii].argmin()
                    shift[ii, 0] = (tv_pxl_id[ii]//w-int(w/2))
                    shift[ii, 1] = (tv_pxl_id[ii] % w-int(w/2))
                    idx = np.int_(shift[levs-1] + int(w/2))
        sh = sh + shift[levs-1]
    elif pxl_conf['type'] == 'line':
        levs = 1
        lsw = pxl_conf['lsw']
        shift = np.zeros([2, 2])
        tv_pxl_id = np.zeros(2, dtype=np.int16)

        tv_v = []
        gmz = zoom(gm, s)
        mk = np.zeros(te.shape, dtype=np.int8)
        mk[int(lsw/2):-int(lsw/2), :] = 1
        mk *= gmz
        for ii in range(lsw):
            tv_v.append(tv_l1_pixel_v2(f, m, mk, gfk, [ii-int(lsw/2), 0]))
        shift[0, 0] = (np.array(tv_v).argmin() - int(lsw/2))
        tv_h = []
        mk = np.zeros(te.shape, dtype=np.int8)
        mk[:, int(lsw/2):-int(lsw/2)] = 1
        mk *= gmz
        for ii in range(lsw):
            tv_h.append(tv_l1_pixel_v2(f, m, mk, gfk, [0, ii-int(lsw/2)]))
        shift[0, 1] = (np.array(tv_h).argmin() - int(lsw/2))
        tv_pxl[0] = np.stack([np.array(tv_v), np.array(tv_h)], axis=0)
        tv_pxl_id[0] = 0
        sh = shift
    elif pxl_conf['type'] == 'al':
        lsw = pxl_conf['lsw']
        for ii in range(levs):
            s = step[ii]
            f = zoom(te, s)
            m = zoom(se, s)
            gmz = zoom(gm, s)
            mk = np.zeros(m.shape, dtype=np.int8)
            mk[int(w*2**(ii-1)):-int(w*2**(ii-1)),
               int(w*2**(ii-1)):-int(w*2**(ii-1))] = 1
            mk *= gmz

            sh = np.array([0, 0])
            for jj in range(ii):
                sh = np.int_(sh + shift[jj]*2**(ii-jj))

            if ii == levs-1:
                tv_v = []
                mk = np.zeros(te.shape, dtype=np.int8)
                mk[int(lsw/2):-int(lsw/2), :] = 1
                mk *= gmz
                for kk in range(lsw):
                    tv_v.append(tv_l1_pixel_v2(f, m, mk, gfk, [kk-int(lsw/2)+sh[0], sh[1]]))
                sh[0] += (np.array(tv_v).argmin() - int(lsw/2))

                tv_h = []
                mk = np.zeros(te.shape, dtype=np.int8)
                mk[:, int(lsw/2):-int(lsw/2)] = 1
                m *= gmz
                for kk in range(lsw):
                    tv_h.append(tv_l1_pixel_v2(f, m, mk, gfk, [sh[0], kk-int(lsw/2)+sh[1]]))
                sh[1] += (np.array(tv_h).argmin() - int(lsw/2))

            tem = np.ndarray([w, w], dtype=np.float32)
            for jj in range(w):
                for kk in range(w):
                    tem[jj, kk] = tv_l1_pixel_v2(f, m, mk, gfk, [jj-int(w/2)+sh[0], kk-int(w/2)+sh[1]])

            tv_pxl[ii] = np.array(tem)
            tv_pxl_id[ii] = tv_pxl[ii].argmin()
            shift[ii, 0] = (tv_pxl_id[ii]//w-int(w/2))
            shift[ii, 1] = (tv_pxl_id[ii] % w-int(w/2))

            if ii == levs - 1:
                idx = np.int_(shift[levs-1] + int(w/2))
                while idx[0] == 0 or idx[0] == w-1 or idx[1] == 0 or idx[1] == w-1:
                    if idx[0] == 0:
                        sh[0] -= int(w/2)
                    elif idx[0] == w-1:
                        sh[0] += int(w/2)
                    if idx[1] == 0:
                        sh[1] -= int(w/2)
                    elif idx[1] == w-1:
                        sh[1] += int(w/2)
                    for jj in range(w):
                        for kk in range(w):
                            tem[jj, kk] = tv_l1_pixel_v2(f, m, mk, gfk, [jj-int(w/2)+sh[0], kk-int(w/2)+sh[1]])

                    tv_pxl[ii] = np.array(tem)
                    tv_pxl_id[ii] = tv_pxl[ii].argmin()
                    shift[ii, 0] = (tv_pxl_id[ii]//w-int(w/2))
                    shift[ii, 1] = (tv_pxl_id[ii] % w-int(w/2))
                    idx = np.int_(shift[levs-1] + int(w/2))
        sh = sh + shift[levs-1]
    elif pxl_conf['type'] == 'la':
        # TODO
        pass

    if sub_conf['use']:
        idx = np.int_(shift[levs-1] + int(w/2))
        if sub_conf['type'] == 'fit':
            sw = int(sub_conf['sp_wz'])
            shift[levs] = tv_l1_subpixel_fit(np.linspace(-sw, sw, 2*sw+1),
                                             np.linspace(-10*sw, 10*sw, 20*sw+1),
                                             [tv_pxl[levs-1].reshape(w, w)
                                              [(idx[0]-sw):(idx[0]+sw+1), idx[1]],
                                              tv_pxl[levs-1].reshape(w, w)
                                              [idx[0], (idx[1]-sw):(idx[1]+sw+1)]])
        elif sub_conf['type'] == 'ana':
            us = int(sub_conf['sp_us'])
            shift[levs] = tv_l1_subpixel_ana(np.linspace(-1, 1, 3),
                                             np.linspace(-us, us, 2*us+1)/us,
                                             [tv_pxl[levs-1].reshape(w, w)
                                              [(idx[0]-1):(idx[0]+2), idx[1]],
                                              tv_pxl[levs-1].reshape(w, w)
                                              [idx[0], (idx[1]-1):(idx[1]+2)]])
    else:
        shift[levs] = [0, 0]
    return tv_pxl, tv_pxl_id, shift, -np.array(sh+shift[levs])


def mrtv_reg_v3(pxl_conf, sub_conf, ps=None, kernel=3, imgs=None):
    """
    Provide a unified interace for using multi-resolution TV-based registration
    algorithm with different configuration options.

    Inputs:
    ______
        pxl_conf: dict; multi-resolution TV minimization configuration at
                  resolution higher than a single pixel; it includes items:
                  'type': 'area', 'line', 'al', 'la'
                  'levs': int
                  'wz': int
                  'lsw': int
        sub_conf: dict; sub-pixel TV minimization configuration; it includes
                  items:
                  'use': boolean; if conduct sub-pixel registration on the end
                  'type': ['ana', 'fit']; which sub-pixel routine to use
                  'sp_wz': int; for 'fit' option; the number of TV points to be
                           used in fitting
                  'sp_us': int; for 'ana' option; up-sampling factor
        ps: optonal; None or a 2-element tuple
        imgs: the pair of the images to be registered

    Outputs:
    _______
        tv_pxl: dict; TV at each search point under levels as the keywords
        tv_pxl_id: the id of the minimum TV in the flatted tv_pxl for each level
        shift: ndarray; shift at each level of resolution
        tot_shift: 2-element tuple; overall displacement between the pair of images
    """
    """
    single process version of mrtv_reg2. Parallelization is implemented
    on the upper level on which the alignments of pairs of images are
    parallelized
    """
    if ps is not None:
        imgs[1] = np.real(np.fft.ifftn(
            fourier_shift(np.fft.fftn(imgs[1]), ps)))

    levs = pxl_conf['levs']
    w = pxl_conf['wz']
    step = {}
    tv_pxl = {}
    tv_pxl_id = np.zeros(levs, dtype=np.int16)
    shift = np.zeros([levs+1, 2], dtype=np.float32)

    for ii in range(levs):
        step[levs-1-ii] = 0.5**ii

    if ((np.array(imgs[0].shape)*0.5**(levs-2) - w) <= 0).any():
        return -1

    if pxl_conf['type'] == 'area':
        for ii in range(levs):
            s = step[ii]
            f = gf(zoom(imgs[0], s), kernel)
            m = gf(zoom(imgs[1], s), kernel)
            mk = np.zeros(m.shape, dtype=np.int8)
            mk[int(w*2**(ii-1)):-int(w*2**(ii-1)),
               int(w*2**(ii-1)):-int(w*2**(ii-1))] = 1

            sh = np.array([0, 0])
            for jj in range(ii):
                sh = np.int_(sh + shift[jj]*2**(ii-jj))

            tem = np.ndarray([w, w], dtype=np.float32)
            for jj in range(w):
                for kk in range(w):
                    tem[jj, kk] = tv_l1_pixel_v3(f, m, mk, [jj-int(w/2)+sh[0], kk-int(w/2)+sh[1]])

            tv_pxl[ii] = np.array(tem)
            tv_pxl_id[ii] = tv_pxl[ii].argmin()
            shift[ii, 0] = (tv_pxl_id[ii]//w-int(w/2))
            shift[ii, 1] = (tv_pxl_id[ii] % w-int(w/2))
            if ii == levs - 1:
                idx = np.int_(shift[levs-1] + int(w/2))
                while idx[0] == 0 or idx[0] == w-1 or idx[1] == 0 or idx[1] == w-1:
                    if idx[0] == 0:
                        sh[0] -= int(w/2)
                    elif idx[0] == w-1:
                        sh[0] += int(w/2)
                    if idx[1] == 0:
                        sh[1] -= int(w/2)
                    elif idx[1] == w-1:
                        sh[1] += int(w/2)
                    for jj in range(w):
                        for kk in range(w):
                            tem[jj, kk] = tv_l1_pixel_v3(f, m, mk, [jj-int(w/2)+sh[0], kk-int(w/2)+sh[1]])

                    tv_pxl[ii] = np.array(tem)
                    tv_pxl_id[ii] = tv_pxl[ii].argmin()
                    shift[ii, 0] = (tv_pxl_id[ii]//w-int(w/2))
                    shift[ii, 1] = (tv_pxl_id[ii] % w-int(w/2))
                    idx = np.int_(shift[levs-1] + int(w/2))
        sh = sh + shift[levs-1]
    elif pxl_conf['type'] == 'line':
        f = gf(imgs[0], kernel)
        m = gf(imgs[1], kernel)
        levs = 1
        lsw = pxl_conf['lsw']
        shift = np.zeros([2, 2])
        tv_pxl_id = np.zeros(2, dtype=np.int16)

        tv_v = []
        mk = np.zeros(imgs[0].shape, dtype=np.int8)
        mk[int(lsw/2):-int(lsw/2), :] = 1
        for ii in range(lsw):
            tv_v.append(tv_l1_pixel_v3(f, m, mk, [ii-int(lsw/2), 0]))
        shift[0, 0] = (np.array(tv_v).argmin() - int(lsw/2))
        tv_h = []
        mk = np.zeros(imgs[0].shape, dtype=np.int8)
        mk[:, int(lsw/2):-int(lsw/2)] = 1
        for ii in range(lsw):
            tv_h.append(tv_l1_pixel_v3(imgs[0], imgs[1], mk, [0, ii-int(lsw/2)]))
        shift[0, 1] = (np.array(tv_h).argmin() - int(lsw/2))
        tv_pxl[0] = np.stack([np.array(tv_v), np.array(tv_h)], axis=0)
        tv_pxl_id[0] = 0
        sh = shift
    elif pxl_conf['type'] == 'al':
        lsw = pxl_conf['lsw']
        for ii in range(levs):
            s = step[ii]
            f = gf(zoom(imgs[0], s), kernel)
            m = gf(zoom(imgs[1], s), kernel)
            mk = np.zeros(m.shape, dtype=np.int8)
            mk[int(w*2**(ii-1)):-int(w*2**(ii-1)),
               int(w*2**(ii-1)):-int(w*2**(ii-1))] = 1

            sh = np.array([0, 0])
            for jj in range(ii):
                sh = np.int_(sh + shift[jj]*2**(ii-jj))

            if ii == levs-1:
                tv_v = []
                mk = np.zeros(imgs[0].shape, dtype=np.int8)
                mk[int(lsw/2):-int(lsw/2), :] = 1
                for kk in range(lsw):
                    tv_v.append(tv_l1_pixel_v3(f, m, mk, [kk-int(lsw/2)+sh[0], sh[1]]))
                sh[0] += (np.array(tv_v).argmin() - int(lsw/2))

                tv_h = []
                mk = np.zeros(imgs[0].shape, dtype=np.int8)
                mk[:, int(lsw/2):-int(lsw/2)] = 1
                for kk in range(lsw):
                    tv_h.append(tv_l1_pixel_v3(f, m, mk, [sh[0], kk-int(lsw/2)+sh[1]]))
                sh[1] += (np.array(tv_h).argmin() - int(lsw/2))

            tem = np.ndarray([w, w], dtype=np.float32)
            for jj in range(w):
                for kk in range(w):
                    tem[jj, kk] = tv_l1_pixel_v3(f, m, mk, [jj-int(w/2)+sh[0], kk-int(w/2)+sh[1]])

            tv_pxl[ii] = np.array(tem)
            tv_pxl_id[ii] = tv_pxl[ii].argmin()
            shift[ii, 0] = (tv_pxl_id[ii]//w-int(w/2))
            shift[ii, 1] = (tv_pxl_id[ii] % w-int(w/2))

            if ii == levs - 1:
                idx = np.int_(shift[levs-1] + int(w/2))
                while idx[0] == 0 or idx[0] == w-1 or idx[1] == 0 or idx[1] == w-1:
                    if idx[0] == 0:
                        sh[0] -= int(w/2)
                    elif idx[0] == w-1:
                        sh[0] += int(w/2)
                    if idx[1] == 0:
                        sh[1] -= int(w/2)
                    elif idx[1] == w-1:
                        sh[1] += int(w/2)
                    for jj in range(w):
                        for kk in range(w):
                            tem[jj, kk] = tv_l1_pixel_v3(f, m, mk, [jj-int(w/2)+sh[0],
                                                       kk-int(w/2)+sh[1]])

                    tv_pxl[ii] = np.array(tem)
                    tv_pxl_id[ii] = tv_pxl[ii].argmin()
                    shift[ii, 0] = (tv_pxl_id[ii]//w-int(w/2))
                    shift[ii, 1] = (tv_pxl_id[ii] % w-int(w/2))
                    idx = np.int_(shift[levs-1] + int(w/2))
        sh = sh + shift[levs-1]
    elif pxl_conf['type'] == 'la':
        # TODO
        pass

    if sub_conf['use']:
        idx = np.int_(shift[levs-1] + int(w/2))
        if sub_conf['type'] == 'fit':
            sw = int(sub_conf['sp_wz'])
            shift[levs] = tv_l1_subpixel_fit(np.linspace(-sw, sw, 2*sw+1),
                                             np.linspace(-10*sw,
                                                         10*sw, 20*sw+1),
                                             [tv_pxl[levs-1].reshape(w, w)
                                              [(idx[0]-sw):(idx[0]+sw+1), idx[1]],
                                              tv_pxl[levs-1].reshape(w, w)
                                              [idx[0], (idx[1]-sw):(idx[1]+sw+1)]])
        elif sub_conf['type'] == 'ana':
            us = int(sub_conf['sp_us'])
            shift[levs] = tv_l1_subpixel_ana(np.linspace(-1, 1, 3),
                                             np.linspace(-us, us, 2*us+1)/us,
                                             [tv_pxl[levs-1].reshape(w, w)
                                              [(idx[0]-1):(idx[0]+2), idx[1]],
                                              tv_pxl[levs-1].reshape(w, w)
                                              [idx[0], (idx[1]-1):(idx[1]+2)]])
    else:
        shift[levs] = [0, 0]
    return tv_pxl, tv_pxl_id, shift, sh+shift[levs]


def mrtv_reg_v4(pxl_conf, sub_conf, ps=None, gfk=3, gm=1, imgs=None, filt=True, norm=True):
    """
    Provide a unified interace for using multi-resolution TV-based registration
    algorithm with different configuration options.

    Inputs:
    ______
        pxl_conf: dict; multi-resolution TV minimization configuration at
                  resolution higher than a single pixel; it includes items:
                  'type': 'area', 'line', 'al', 'la'
                  'levs': int
                  'wz': int
                  'lsw': int
        sub_conf: dict; sub-pixel TV minimization configuration; it includes
                  items:
                  'use': boolean; if conduct sub-pixel registration on the end
                  'type': ['ana', 'fit']; which sub-pixel routine to use
                  'sp_wz': int; for 'fit' option; the number of TV points to be
                           used in fitting
                  'sp_us': int; for 'ana' option; up-sampling factor
        ps: optonal; None or a 2-element tuple
        imgs: the pair of the images to be registered

    Outputs:
    _______
        tv_pxl: dict; TV at each search point under levels as the keywords
        tv_pxl_id: the id of the minimum TV in the flatted tv_pxl for each level
        shift: ndarray; shift at each level of resolution
        tot_shift: 2-element tuple; overall displacement between the pair of images
    """
    """
    single process version of mrtv_reg2. Parallelization is implemented
    on the upper level on which the alignments of pairs of images are
    parallelized
    """
    if ps is not None:
        imgs[1] = np.real(np.fft.ifftn(
            fourier_shift(np.fft.fftn(imgs[1]), ps)))

    levs = pxl_conf['levs']
    w = pxl_conf['wz']
    step = {}
    tv_pxl = {}
    tv_pxl_id = np.zeros(levs, dtype=np.int16)
    shift = np.zeros([levs+1, 2], dtype=np.float32)

    for ii in range(levs):
        step[levs-1-ii] = 0.5**ii

    if ((np.array(imgs[0].shape)*0.5**(levs-2) - w) <= 0).any():
        return -1

    if not isinstance(gm, np.ndarray):
        gm = np.ndarray(imgs[0].shape, dtype=np.int8)

    if pxl_conf['type'] == 'area':
        for ii in range(levs):
            s = step[ii]
            f = zoom(imgs[0], s)
            m = zoom(imgs[1], s)
            gmz = zoom(gm, s)
            mk = np.zeros(m.shape, dtype=np.int8)
            mk[int(w*2**(ii-1)):-int(w*2**(ii-1)),
               int(w*2**(ii-1)):-int(w*2**(ii-1))] = 1
            mk *= gmz

            sh = np.array([0, 0])
            for jj in range(ii):
                sh = np.int_(sh + shift[jj]*2**(ii-jj))

            tem = np.ndarray([w, w], dtype=np.float32)
            for jj in range(w):
                for kk in range(w):
                    tem[jj, kk] = tv_l1_pixel_v4(f, m, mk,
                                                 [jj-int(w/2)+sh[0],
                                                 kk-int(w/2)+sh[1]],
                                                 kernel=gfk, filt=filt,
                                                 norm=norm)

            tv_pxl[ii] = np.array(tem)
            tv_pxl_id[ii] = tv_pxl[ii].argmin()
            shift[ii, 0] = (tv_pxl_id[ii]//w-int(w/2))
            shift[ii, 1] = (tv_pxl_id[ii] % w-int(w/2))
            if ii == levs - 1:
                idx = np.int_(shift[levs-1] + int(w/2))
                while idx[0] == 0 or idx[0] == w-1 or idx[1] == 0 or idx[1] == w-1:
                    if idx[0] == 0:
                        sh[0] -= int(w/2)
                    elif idx[0] == w-1:
                        sh[0] += int(w/2)
                    if idx[1] == 0:
                        sh[1] -= int(w/2)
                    elif idx[1] == w-1:
                        sh[1] += int(w/2)
                    for jj in range(w):
                        for kk in range(w):
                            tem[jj, kk] = tv_l1_pixel_v4(f, m, mk,
                                                         [jj-int(w/2)+sh[0],
                                                          kk-int(w/2)+sh[1]],
                                                         kernel=gfk, filt=filt,
                                                         norm=norm)

                    tv_pxl[ii] = np.array(tem)
                    tv_pxl_id[ii] = tv_pxl[ii].argmin()
                    shift[ii, 0] = (tv_pxl_id[ii]//w-int(w/2))
                    shift[ii, 1] = (tv_pxl_id[ii] % w-int(w/2))
                    idx = np.int_(shift[levs-1] + int(w/2))
        sh = sh + shift[levs-1]
    elif pxl_conf['type'] == 'line':
        levs = 1
        lsw = pxl_conf['lsw']
        shift = np.zeros([2, 2])
        tv_pxl_id = np.zeros(2, dtype=np.int16)

        tv_v = []
        mk = np.zeros(imgs[0].shape, dtype=np.int8)
        mk[int(lsw/2):-int(lsw/2), :] = 1
        mk *= gm
        for ii in range(lsw):
            tv_v.append(tv_l1_pixel_v4(imgs[0], imgs[1], mk,
                                       [ii-int(lsw/2), 0],
                                       kernel=gfk, filt=filt,
                                       norm=norm))
        shift[0, 0] = (np.array(tv_v).argmin() - int(lsw/2))
        tv_h = []
        mk = np.zeros(imgs[0].shape, dtype=np.int8)
        mk[:, int(lsw/2):-int(lsw/2)] = 1
        mk *= gm
        for ii in range(lsw):
            tv_h.append(tv_l1_pixel_v4(imgs[0], imgs[1], mk,
                                       [0, ii-int(lsw/2)],
                                       kernel=gfk, filt=filt,
                                       norm=norm))
        shift[0, 1] = (np.array(tv_h).argmin() - int(lsw/2))
        tv_pxl[0] = np.stack([np.array(tv_v), np.array(tv_h)], axis=0)
        tv_pxl_id[0] = 0
        sh = shift
    elif pxl_conf['type'] == 'al':
        lsw = pxl_conf['lsw']
        for ii in range(levs):
            s = step[ii]
            f = zoom(imgs[0], s)
            m = zoom(imgs[1], s)
            gmz = zoom(gm, s)

            sh = np.array([0, 0])
            for jj in range(ii):
                sh = np.int_(sh + shift[jj]*2**(ii-jj))

            if ii == levs-1:
                tv_v = []
                mk = np.zeros(imgs[0].shape, dtype=np.int8)
                mk[int(lsw/2):-int(lsw/2), :] = 1
                mk *= gmz
                for kk in range(lsw):
                    tv_v.append(tv_l1_pixel_v4(f, m, mk,
                                               [kk-int(lsw/2)+sh[0], sh[1]],
                                               kernel=gfk, filt=filt,
                                               norm=norm))
                sh[0] += (np.array(tv_v).argmin() - int(lsw/2))

                tv_h = []
                mk = np.zeros(imgs[0].shape, dtype=np.int8)
                mk[:, int(lsw/2):-int(lsw/2)] = 1
                mk *= gmz
                for kk in range(lsw):
                    tv_h.append(tv_l1_pixel_v4(f, m, mk,
                                               [sh[0], kk-int(lsw/2)+sh[1]],
                                               kernel=gfk, filt=filt,
                                               norm=norm))
                sh[1] += (np.array(tv_h).argmin() - int(lsw/2))

            tem = np.ndarray([w, w], dtype=np.float32)
            mk = np.zeros(m.shape, dtype=np.int8)
            mk[int(w * 2 ** (ii - 1)):-int(w * 2 ** (ii - 1)),
            int(w * 2 ** (ii - 1)):-int(w * 2 ** (ii - 1))] = 1
            mk *= gmz
            for jj in range(w):
                for kk in range(w):
                    tem[jj, kk] = tv_l1_pixel_v4(f, m, mk,
                                                 [jj-int(w/2)+sh[0],
                                                  kk-int(w/2)+sh[1]],
                                                 kernel=gfk, filt=filt,
                                                 norm=norm)

            tv_pxl[ii] = np.array(tem)
            tv_pxl_id[ii] = tv_pxl[ii].argmin()
            shift[ii, 0] = (tv_pxl_id[ii]//w-int(w/2))
            shift[ii, 1] = (tv_pxl_id[ii] % w-int(w/2))

            if ii == levs - 1:
                idx = np.int_(shift[levs-1] + int(w/2))
                while idx[0] == 0 or idx[0] == w-1 or idx[1] == 0 or idx[1] == w-1:
                    if idx[0] == 0:
                        sh[0] -= int(w/2)
                    elif idx[0] == w-1:
                        sh[0] += int(w/2)
                    if idx[1] == 0:
                        sh[1] -= int(w/2)
                    elif idx[1] == w-1:
                        sh[1] += int(w/2)
                    for jj in range(w):
                        for kk in range(w):
                            tem[jj, kk] = tv_l1_pixel_v4(f, m, mk,
                                                         [jj-int(w/2)+sh[0],
                                                          kk-int(w/2)+sh[1]],
                                                         kernel=gfk, filt=filt,
                                                         norm=norm)

                    tv_pxl[ii] = np.array(tem)
                    tv_pxl_id[ii] = tv_pxl[ii].argmin()
                    shift[ii, 0] = (tv_pxl_id[ii]//w-int(w/2))
                    shift[ii, 1] = (tv_pxl_id[ii] % w-int(w/2))
                    idx = np.int_(shift[levs-1] + int(w/2))
        sh = sh + shift[levs-1]
    elif pxl_conf['type'] == 'la':
        # TODO
        pass

    if sub_conf['use']:
        idx = np.int_(shift[levs-1] + int(w/2))
        if sub_conf['type'] == 'fit':
            sw = int(sub_conf['sp_wz'])
            shift[levs] = tv_l1_subpixel_fit(np.linspace(-sw, sw, 2*sw+1),
                                             np.linspace(-10*sw,
                                                         10*sw, 20*sw+1),
                                             [tv_pxl[levs-1].reshape(w, w)
                                              [(idx[0]-sw):(idx[0]+sw+1), idx[1]],
                                              tv_pxl[levs-1].reshape(w, w)
                                              [idx[0], (idx[1]-sw):(idx[1]+sw+1)]])
        elif sub_conf['type'] == 'ana':
            us = int(sub_conf['sp_us'])
            shift[levs] = tv_l1_subpixel_ana(np.linspace(-1, 1, 3),
                                             np.linspace(-us, us, 2*us+1)/us,
                                             [tv_pxl[levs-1].reshape(w, w)
                                              [(idx[0]-1):(idx[0]+2), idx[1]],
                                              tv_pxl[levs-1].reshape(w, w)
                                              [idx[0], (idx[1]-1):(idx[1]+2)]])
    else:
        shift[levs] = [0, 0]
    return tv_pxl, tv_pxl_id, shift, sh+shift[levs]


def mrtv_mpc_combo_reg(fixed_img, img, us=100, reference_mask=None,
                       overlap_ratio=0.3, levs=4, wz=10, sp_wz=20,
                       sp_step=0.2):
    shift = np.zeros([levs+1, 2])
    if reference_mask is not None:
        shift[0] = phase_cross_correlation(fixed_img, img, upsample_factor=us,
                                           reference_mask=reference_mask,
                                           overlap_ratio=overlap_ratio)
    else:
        shift[0], _, _ = phase_cross_correlation(fixed_img, img,
                                                 upsample_factor=us,
                                                 overlap_ratio=overlap_ratio)

    tv1_pxl, tv1_pxl_id, shift[1:], ss = mrtv_reg2(fixed_img, img, levs=levs,
                                                   wz=wz, sp_wz=sp_wz,
                                                   sp_step=sp_step, ps=shift[0])

    print(time.asctime())
    return tv1_pxl, tv1_pxl_id, shift, shift[0]+ss


@msgit(wd=100, fill='-')
def mrtv_ls_combo_reg(ls_w=100, levs=2, wz=10, sp_wz=8, sp_step=0.5, kernel=3, imgs=None):
    """
    line search combined with two-level multi-resolution search
    """
    shift = np.zeros([3, 2])
    tv1_pxl_id = np.zeros(2, dtype=np.int16)
    tv1_pxl = {}
    tv_v = []
    for ii in range(ls_w):
        tv_v.append(tv_l1_pixel(imgs[0], imgs[1], 1, [2*ii-ls_w, 0],
                                kernel=kernel))
    shift[0, 0] = 2*(np.array(tv_v).argmin() - int(ls_w/2))

    tv_h = []
    for ii in range(ls_w):
        tv_h.append(tv_l1_pixel(imgs[0], imgs[1], 1, [0, 2*ii-ls_w],
                                kernel=kernel))
    shift[0, 1] = 2*(np.array(tv_h).argmin() - int(ls_w/2))

    tv1_pxl, tv1_pxl_id, shift[1:], ss = mrtv_reg3(levs=levs, wz=wz, sp_wz=sp_wz,
                                                   sp_step=sp_step, imgs=imgs,
                                                   ps=shift[0])

    print(time.asctime())
    return tv1_pxl, tv1_pxl_id, shift, shift[0]+ss


@msgit(wd=100, fill='-')
def mrtv_ls_combo_reg2(ls_w=100, levs=2, wz=10, sp_wz=8, sp_step=0.5, kernel=3, imgs=None):
    """
    line search combined with two-level multi-resolution search
    """
    shift = np.zeros([3, 2])
    tv1_pxl_id = np.zeros(2, dtype=np.int16)
    tv1_pxl = {}
    tv_v = []
    for ii in range(ls_w):
        tv_v.append(tv_l1_pixel(imgs[0], imgs[1], 1, [ii-int(ls_w/2), 0],
                                kernel=kernel))
    shift[0, 0] = (np.array(tv_v).argmin() - int(ls_w/2))

    tv_h = []
    for ii in range(ls_w):
        tv_h.append(tv_l1_pixel(imgs[0], imgs[1], 1, [0, ii-int(ls_w/2)],
                                kernel=kernel))
    shift[0, 1] = (np.array(tv_h).argmin() - int(ls_w/2))

    tv1_pxl, tv1_pxl_id, shift[1:], ss = mrtv_reg4(levs=levs, wz=wz, sp_wz=sp_wz,
                                                   sp_step=sp_step, imgs=imgs,
                                                   ps=shift[0])

    print(time.asctime())
    return tv1_pxl, tv1_pxl_id, shift, shift[0]+ss


def shift_img(img_shift):
    img = img_shift[0]
    shift = img_shift[1]
    return np.real(np.fft.ifftn(fourier_shift(np.fft.fftn(img), shift)))


def sr_transform_img(sr, data):
    return sr.transform(data[0], tmat=data[1])


def mp_shift_img(imgs, shifts, stype='trans', sr=None, axis=0):
    if stype.upper() == 'SR':
        if axis == 0:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(sr_transform_img, sr),
                               [[imgs[ii, ...], shifts[ii]] for ii
                                in range(imgs.shape[0])])
            pool.close()
            pool.join()
            return np.array(rlt).astype(np.float32)
        elif axis == 1:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(sr_transform_img, sr),
                               [[imgs[:, ii, :], shifts[ii]] for ii
                                in range(imgs.shape[1])])
            pool.close()
            pool.join()
            return np.swapaxes(np.array(rlt).astype(np.float32), 0, 1)
        elif axis == 2:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(sr_transform_img, sr),
                               [[imgs[:, :, ii], shifts[ii]] for ii
                                in range(imgs.shape[2])])
            pool.close()
            pool.join()
            return np.swapaxes(np.array(rlt).astype(np.float32), 0, 2)
    else:
        if axis == 0:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(shift_img,
                               [[imgs[ii, ...], shifts[ii]] for ii
                                in range(imgs.shape[0])])
            pool.close()
            pool.join()
            return np.array(rlt).astype(np.float32)
        elif axis == 1:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(shift_img,
                               [[imgs[:, ii, :], shifts[ii]] for ii
                                in range(imgs.shape[1])])
            pool.close()
            pool.join()
            return np.swapaxes(np.array(rlt).astype(np.float32), 0, 1)
        elif axis == 2:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(shift_img,
                               [[imgs[:, :, ii], shifts[ii]] for ii
                                in range(imgs.shape[2])])
            pool.close()
            pool.join()
            return np.swapaxes(np.array(rlt).astype(np.float32), 0, 2)


def pc_reg(upsample_factor, imgs):
    shift, _, _ = phase_cross_correlation(
        imgs[0], imgs[1], upsample_factor=upsample_factor)
    return shift


def mpc_reg(reference_mask, overlap_ratio, imgs):
    shift = phase_cross_correlation(imgs[0], imgs[1], reference_mask=reference_mask,
                                    overlap_ratio=overlap_ratio)
    return shift


def sr_reg(sr, imgs):
    shift = sr.register(imgs[0], imgs[1])
    return shift


def reg_stack(imgs, method, *args, axis=0, filter=None, **kwargs):
    shift = []
    if method.upper() == 'MRTV_REG':
        if filter is None:
            if axis == 0:
                with mp.Pool(N_CPU) as pool:
                    rlt = pool.map(partial(mrtv_reg, *args, **kwargs),
                                   [[imgs[ii], imgs[ii+1]] for ii
                                    in range(imgs.shape[0]-1)])
            elif axis == 1:
                with mp.Pool(N_CPU) as pool:
                    rlt = pool.map(partial(mrtv_reg, *args, **kwargs),
                                   [[imgs[:, ii, :], imgs[:, ii+1, :]] for ii
                                    in range(imgs.shape[1]-1)])
            elif axis == 2:
                with mp.Pool(N_CPU) as pool:
                    rlt = pool.map(partial(mrtv_reg, *args, **kwargs),
                                   [[imgs[:, :, ii], imgs[:, :, ii+1]] for ii
                                    in range(imgs.shape[2]-1)])
            pool.close()
            pool.join()
            shift = np.vstack(np.array(rlt, dtype=object)[:, -1])
        else:
            flt_func = filter['func']
            flt_parm = filter['params']
            if axis == 0:
                with mp.Pool(N_CPU) as pool:
                    rlt = pool.map(partial(mrtv_reg, *args, **kwargs),
                                   [[flt_func(imgs[ii], *flt_parm), flt_func(imgs[ii+1], *flt_parm)] for ii
                                    in range(imgs.shape[0]-1)])
            elif axis == 1:
                with mp.Pool(N_CPU) as pool:
                    rlt = pool.map(partial(mrtv_reg, *args, **kwargs),
                                   [[flt_parm(imgs[:, ii, :], *flt_parm), flt_func(imgs[:, ii+1, :], *flt_parm)] for ii
                                    in range(imgs.shape[1]-1)])
            elif axis == 2:
                with mp.Pool(N_CPU) as pool:
                    rlt = pool.map(partial(mrtv_reg, *args, **kwargs),
                                   [[flt_parm(imgs[:, :, ii], *flt_parm), flt_func(imgs[:, :, ii+1], *flt_parm)] for ii
                                    in range(imgs.shape[2]-1)])
            pool.close()
            pool.join()
            shift = np.vstack(np.array(rlt, dtype=object)[:, -1])
    elif method.upper() == 'PC':
        if axis == 0:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(pc_reg, *args, **kwargs),
                               [[imgs[ii], imgs[ii+1]] for ii
                                in range(imgs.shape[0]-1)])
        elif axis == 1:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(pc_reg, *args, **kwargs),
                               [[imgs[:, ii, :], imgs[:, ii+1, :]] for ii
                                in range(imgs.shape[1]-1)])
        elif axis == 2:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(pc_reg, *args, **kwargs),
                               [[imgs[:, :, ii], imgs[:, :, ii+1]] for ii
                                in range(imgs.shape[2]-1)])
        pool.close()
        pool.join()
        shift = np.array(rlt)
    elif method.upper() == 'MPC':
        if axis == 0:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(mpc_reg, *args, **kwargs),
                               [[imgs[ii], imgs[ii+1]] for ii
                                in range(imgs.shape[0]-1)])
        elif axis == 1:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(mpc_reg, *args, **kwargs),
                               [[imgs[:, ii, :], imgs[:, ii+1, :]] for ii
                                in range(imgs.shape[1]-1)])
        elif axis == 2:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(mpc_reg, *args, **kwargs),
                               [[imgs[:, :, ii], imgs[:, :, ii+1]] for ii
                                in range(imgs.shape[2]-1)])
        pool.close()
        pool.join()
        shift = np.array(rlt)
    elif method.upper() == 'SR':
        if axis == 0:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(sr_reg, *args, **kwargs),
                               [[imgs[ii], imgs[ii+1]] for ii
                                in range(imgs.shape[0]-1)])
        elif axis == 1:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(sr_reg, *args, **kwargs),
                               [[imgs[:, ii, :], imgs[:, ii+1, :]] for ii
                                in range(imgs.shape[1]-1)])
        elif axis == 2:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(sr_reg, *args, **kwargs),
                               [[imgs[:, :, ii], imgs[:, :, ii+1]] for ii
                                in range(imgs.shape[2]-1)])
        pool.close()
        pool.join()
        shift = np.array(rlt)
    return shift


def reg_stacks(refs, movs, method, *args, axis=0, filter=None, **kwargs):
    shift = []
    if method.upper() == 'MRTV_REG':
        if filter is None:
            if axis == 0:
                with mp.Pool(N_CPU) as pool:
                    rlt = pool.map(partial(mrtv_reg, *args, **kwargs),
                                   [[refs[ii], movs[ii]] for ii
                                    in range(refs.shape[0])])
            elif axis == 1:
                with mp.Pool(N_CPU) as pool:
                    rlt = pool.map(partial(mrtv_reg, *args, **kwargs),
                                   [[refs[:, ii, :], movs[:, ii, :]] for ii
                                    in range(refs.shape[1])])
            elif axis == 2:
                with mp.Pool(N_CPU) as pool:
                    rlt = pool.map(partial(mrtv_reg, *args, **kwargs),
                                   [[refs[:, :, ii], movs[:, :, ii]] for ii
                                    in range(refs.shape[2])])
        else:
            flt_func = filter['func']
            flt_parm = filter['params']
            if axis == 0:
                with mp.Pool(N_CPU) as pool:
                    rlt = pool.map(partial(mrtv_reg, *args, **kwargs),
                                   [[flt_func(refs[ii], *flt_parm), flt_func(movs[ii], *flt_parm)] for ii
                                    in range(refs.shape[0])])
            elif axis == 1:
                with mp.Pool(N_CPU) as pool:
                    rlt = pool.map(partial(mrtv_reg, *args, **kwargs),
                                   [[flt_func(refs[:, ii, :], *flt_parm), flt_func(movs[:, ii, :], *flt_parm)] for ii
                                    in range(refs.shape[1])])
            elif axis == 2:
                with mp.Pool(N_CPU) as pool:
                    rlt = pool.map(partial(mrtv_reg, *args, **kwargs),
                                   [[flt_func(refs[:, :, ii], *flt_parm), flt_func(movs[:, :, ii], *flt_parm)] for ii
                                    in range(refs.shape[2])])
        pool.close()
        pool.join()
        shift = np.vstack(np.array(rlt, dtype=object)[:, -1])
    elif method.upper() == 'SR':
        if axis == 0:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(sr_reg, *args, **kwargs),
                               [[refs[ii], movs[ii]] for ii
                                in range(refs.shape[0])])
        elif axis == 1:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(sr_reg, *args, **kwargs),
                               [[refs[:, ii, :], movs[:, ii, :]] for ii
                                in range(refs.shape[1])])
        elif axis == 2:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(sr_reg, *args, **kwargs),
                               [[refs[:, :, ii], movs[:, :, ii]] for ii
                                in range(refs.shape[2])])
        pool.close()
        pool.join()
        shift = np.array(rlt)
    elif method.upper() == 'PC':
        if axis == 0:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(pc_reg, *args, **kwargs),
                               [[refs[ii], movs[ii]] for ii
                                in range(refs.shape[0])])
        elif axis == 1:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(pc_reg, *args, **kwargs),
                               [[refs[:, ii, :], movs[:, ii, :]] for ii
                                in range(refs.shape[1])])
        elif axis == 2:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(pc_reg, *args, **kwargs),
                               [[refs[:, :, ii], movs[:, :, ii]] for ii
                                in range(refs.shape[2])])
        pool.close()
        pool.join()
        shift = np.array(rlt)
    elif method.upper() == 'MPC':
        if axis == 0:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(mpc_reg, *args, **kwargs),
                               [[refs[ii], movs[ii]] for ii
                                in range(refs.shape[0])])
        elif axis == 1:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(mpc_reg, *args, **kwargs),
                               [[refs[:, ii, :], movs[:, ii, :]] for ii
                                in range(refs.shape[1])])
        elif axis == 2:
            with mp.Pool(N_CPU) as pool:
                rlt = pool.map(partial(mpc_reg, *args, **kwargs),
                               [[refs[:, :, ii], movs[:, :, ii]] for ii
                                in range(refs.shape[2])])
        pool.close()
        pool.join()
        shift = np.array(rlt)
    return shift



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

def user_tomo_reader(*args, **kwargs):
    pass

def user_xanes2D_reader(*args, **kwargs):
    pass

def user_xanes3D_reader(*args, **kwargs):
    pass

def user_tomo_info_reader(*args, **kwargs):
    pass


"""Adpated from larch"""

"""Basic model line shapes and distribution functions."""

from numpy import (arctan, cos, exp, finfo, float64, isnan, log, pi, real, sin,
                   sqrt, where, polyval)
from scipy.special import erf, erfc
from scipy.special import gamma as gamfcn
from scipy.special import wofz

log2 = log(2)
s2pi = sqrt(2*pi)
spi = sqrt(pi)
s2 = sqrt(2.0)
tiny = finfo(float64).eps

functions = ('gaussian', 'lorentzian', 'voigt', 'pvoigt', 'moffat', 'pearson7',
             'breit_wigner', 'damped_oscillator', 'dho', 'logistic', 'lognormal',
             'students_t', 'expgaussian', 'doniach', 'donaich', 'skewed_gaussian',
             'skewed_voigt', 'thermal_distribution', 'step', 'rectangle',
             'exponential', 'powerlaw', 'linear', 'parabolic', 'sine',
             'expsine', 'split_lorentzian')


def gaussian(x, amplitude=1.0, center=0.0, sigma=1.0):
    """Return a 1-dimensional Gaussian function.

    gaussian(x, amplitude, center, sigma) =
        (amplitude/(s2pi*sigma)) * exp(-(1.0*x-center)**2 / (2*sigma**2))

    """
    return ((amplitude/(max(tiny, s2pi*sigma)))
            * exp(-(1.0*x-center)**2 / max(tiny, (2*sigma**2))))


def lorentzian(x, amplitude=1.0, center=0.0, sigma=1.0):
    """Return a 1-dimensional Lorentzian function.

    lorentzian(x, amplitude, center, sigma) =
        (amplitude/(1 + ((1.0*x-center)/sigma)**2)) / (pi*sigma)

    """
    return ((amplitude/(1 + ((1.0*x-center)/max(tiny, sigma))**2))
            / max(tiny, (pi*sigma)))


def split_lorentzian(x, amplitude=1.0, center=0.0, sigma=1.0, sigma_r=1.0):
    """Return a 1-dimensional piecewise Lorentzian function.

    Split means that width of the function is different between
    left and right slope of the function. The peak height is calculated
    from the condition that the integral from ``-.inf`` to ``+.inf`` is equal
    to ``amplitude``.

    split_lorentzian(x, amplitude, center, sigma, sigma_r) =
        [2*amplitude / (pi* (sigma + sigma_r)] *
         {  sigma**2   * (x<center)  / [sigma**2  + (x - center)**2]
          + sigma_r**2 * (x>=center) / [sigma_r**2+ (x - center)**2] }

    """
    s = max(tiny, sigma)
    r = max(tiny, sigma_r)
    s2 = s*s
    r2 = r*r
    xc2 = (x-center)**2
    amp = 2*amplitude/(pi*(s+r))
    return amp*(s2*(x < center)/(s2+xc2) + r2*(x >= center)/(r2+xc2))


def voigt(x, amplitude=1.0, center=0.0, sigma=1.0, gamma=None):
    """Return a 1-dimensional Voigt function.

    voigt(x, amplitude, center, sigma, gamma) =
        amplitude*wofz(z).real / (sigma*s2pi)

    see https://en.wikipedia.org/wiki/Voigt_profile

    """
    if gamma is None:
        gamma = sigma
    z = (x-center + 1j*gamma) / max(tiny, (sigma*s2))
    return amplitude*wofz(z).real / max(tiny, (sigma*s2pi))


def pvoigt(x, amplitude=1.0, center=0.0, sigma=1.0, fraction=0.5):
    """Return a 1-dimensional pseudo-Voigt function.

    pvoigt(x, amplitude, center, sigma, fraction) =
       amplitude*(1-fraction)*gaussian(x, center, sigma_g) +
       amplitude*fraction*lorentzian(x, center, sigma)

    where sigma_g (the sigma for the Gaussian component) is

        sigma_g = sigma / sqrt(2*log(2)) ~= sigma / 1.17741

    so that the Gaussian and Lorentzian components have the
    same FWHM of 2*sigma.

    """
    sigma_g = sigma / sqrt(2*log2)
    return ((1-fraction)*gaussian(x, amplitude, center, sigma_g) +
            fraction*lorentzian(x, amplitude, center, sigma))


def moffat(x, amplitude=1, center=0., sigma=1, beta=1.):
    """Return a 1-dimensional Moffat function.

    moffat(x, amplitude, center, sigma, beta) =
        amplitude / (((x - center)/sigma)**2 + 1)**beta

    """
    return amplitude / (((x - center)/max(tiny, sigma))**2 + 1)**beta


def pearson7(x, amplitude=1.0, center=0.0, sigma=1.0, expon=1.0):
    """Return a Pearson7 lineshape.

    Using the wikipedia definition:
    pearson7(x, center, sigma, expon) =
        amplitude*(1+arg**2)**(-expon)/(sigma*beta(expon-0.5, 0.5))

    where arg = (x-center)/sigma
    and beta() is the beta function.

    """
    expon = max(tiny, expon)
    arg = (x-center)/max(tiny, sigma)
    scale = amplitude * gamfcn(expon)/(gamfcn(0.5)*gamfcn(expon-0.5))
    return scale*(1+arg**2)**(-expon)/max(tiny, sigma)


def breit_wigner(x, amplitude=1.0, center=0.0, sigma=1.0, q=1.0):
    """Return a Breit-Wigner-Fano lineshape.

    breit_wigner(x, amplitude, center, sigma, q) =
        amplitude*(q*sigma/2 + x - center)**2 /
            ( (sigma/2)**2 + (x - center)**2 )

    """
    gam = sigma/2.0
    return amplitude*(q*gam + x - center)**2 / (gam*gam + (x-center)**2)


def damped_oscillator(x, amplitude=1.0, center=1., sigma=0.1):
    """Return the amplitude for a damped harmonic oscillator.

    damped_oscillator(x, amplitude, center, sigma) =
        amplitude/sqrt( (1.0 - (x/center)**2)**2 + (2*sigma*x/center)**2))

    """
    center = max(tiny, abs(center))
    return amplitude/sqrt((1.0 - (x/center)**2)**2 + (2*sigma*x/center)**2)


def dho(x, amplitude=1., center=0., sigma=1., gamma=1.0):
    """Return a Damped Harmonic Oscillator.

    Similar to version from PAN
    dho(x, amplitude, center, sigma, gamma) =
        amplitude*sigma*pi * (lm - lp) / (1.0 - exp(-x/gamma))

    where
        lm(x, center, sigma) = 1.0 / ((x-center)**2 + sigma**2)
        lp(x, center, sigma) = 1.0 / ((x+center)**2 + sigma**2)

    """
    bose = (1.0 - exp(-x/max(tiny, gamma)))
    if isinstance(bose, (int, float)):
        bose = max(tiny, bose)
    else:
        bose[where(isnan(bose))] = tiny
        bose[where(bose <= tiny)] = tiny

    lm = 1.0/((x-center)**2 + sigma**2)
    lp = 1.0/((x+center)**2 + sigma**2)
    return amplitude*sigma/pi*(lm - lp)/bose


def logistic(x, amplitude=1., center=0., sigma=1.):
    """Return a Logistic lineshape (yet another sigmoidal curve).

    logistic(x, amplitude, center, sigma) =
        = amplitude*(1.  - 1. / (1 + exp((x-center)/sigma)))

    """
    return amplitude*(1. - 1./(1. + exp((x-center)/max(tiny, sigma))))


def lognormal(x, amplitude=1.0, center=0., sigma=1):
    """Return a log-normal function.

    lognormal(x, amplitude, center, sigma)
        = (amplitude/(x*sigma*s2pi)) * exp(-(ln(x) - center)**2/ (2* sigma**2))

    """
    if isinstance(x, (int, float)):
        x = max(tiny, x)
    else:
        x[where(x <= tiny)] = tiny
    return ((amplitude/(x*max(tiny, sigma*s2pi))) *
            exp(-(log(x)-center)**2 / max(tiny, (2*sigma**2))))


def students_t(x, amplitude=1.0, center=0.0, sigma=1.0):
    """Return Student's t distribution.

    students_t(x, amplitude, center, sigma) =

             gamma((sigma+1)/2)
        ---------------------------- * (1 + (x-center)**2/sigma)^(-(sigma+1)/2)
        sqrt(sigma*pi)gamma(sigma/2)

    """
    s1 = (sigma+1)/2.0
    denom = max(tiny, (sqrt(sigma*pi)*gamfcn(max(tiny, sigma/2))))
    return amplitude*(1 + (x-center)**2/max(tiny, sigma))**(-s1) * gamfcn(s1) / denom


def expgaussian(x, amplitude=1, center=0, sigma=1.0, gamma=1.0):
    """Return an exponentially modified Gaussian.

    expgaussian(x, amplitude, center, sigma, gamma)
        = amplitude * (gamma/2) *
          exp[center*gamma + (gamma*sigma)**2/2 - gamma*x] *
          erfc[(center + gamma*sigma**2 - x)/(sqrt(2)*sigma)]

    https://en.wikipedia.org/wiki/Exponentially_modified_Gaussian_distribution

    """
    gss = gamma*sigma*sigma
    arg1 = gamma*(center + gss/2.0 - x)
    arg2 = (center + gss - x)/max(tiny, (s2*sigma))
    return amplitude*(gamma/2) * exp(arg1) * erfc(arg2)


def doniach(x, amplitude=1.0, center=0, sigma=1.0, gamma=0.0):
    """Return a Doniach Sunjic asymmetric lineshape, used for photo-emission.

    doniach(x, amplitude, center, sigma, gamma) =
        amplitude / sigma^(1-gamma) *
        cos(pi*gamma/2 + (1-gamma) arctan((x-center)/sigma) /
        (sigma**2 + (x-center)**2)**[(1-gamma)/2]

    see http://www.casaxps.com/help_manual/line_shapes.htm

    """
    arg = (x-center)/max(tiny, sigma)
    gm1 = (1.0 - gamma)
    scale = amplitude/max(tiny, (sigma**gm1))
    return scale*cos(pi*gamma/2 + gm1*arctan(arg))/(1 + arg**2)**(gm1/2)


donaich = doniach   # for back-compat


def skewed_gaussian(x, amplitude=1.0, center=0.0, sigma=1.0, gamma=0.0):
    """Return a Gaussian lineshape, skewed with error function.

    Equal to: gaussian(x, center, sigma)*(1+erf(beta*(x-center)))

    with beta = gamma/(sigma*sqrt(2))

    with  gamma < 0:  tail to low value of centroid
          gamma > 0:  tail to high value of centroid

    see https://en.wikipedia.org/wiki/Skew_normal_distribution

    """
    asym = 1 + erf(gamma*(x-center)/max(tiny, (s2*sigma)))
    return asym * gaussian(x, amplitude, center, sigma)


def skewed_voigt(x, amplitude=1.0, center=0.0, sigma=1.0, gamma=None, skew=0.0):
    """Return a Voigt lineshape, skewed with error function.

    useful for ad-hoc Compton scatter profile

    with beta = skew/(sigma*sqrt(2))
    = voigt(x, center, sigma, gamma)*(1+erf(beta*(x-center)))

    skew < 0:  tail to low value of centroid
    skew > 0:  tail to high value of centroid

    see https://en.wikipedia.org/wiki/Skew_normal_distribution

    """
    beta = skew/max(tiny, (s2*sigma))
    asym = 1 + erf(beta*(x-center))
    return asym * voigt(x, amplitude, center, sigma, gamma=gamma)


def sine(x, amplitude=1.0, frequency=1.0, shift=0.0):
    """Return a sinusoidal function.

    sine(x, amplitude, frequency, shift):
       = amplitude * sin(x*frequency + shift)

    """
    return amplitude*sin(x*frequency + shift)


def expsine(x, amplitude=1.0, frequency=1.0, shift=0.0, decay=0.0):
    """Return an exponentially decaying sinusoidal function.

    expsine(x, amplitude, frequency, shift,  decay):
       = amplitude * sin(x*frequency + shift) * exp(-x*decay)

    """
    return amplitude*sin(x*frequency + shift) * exp(-x*decay)


def thermal_distribution(x, amplitude=1.0, center=0.0, kt=1.0, form='bose'):
    """Return a thermal distribution function.

    form = 'bose' (default) is the Bose-Einstein distribution
    thermal_distribution(x, amplitude=1.0, center=0.0, kt=1.0):
       = 1/(amplitude*exp((x - center)/kt) - 1)
    form = 'maxwell' is the Maxwell-Boltzmann distribution
    thermal_distribution(x, amplitude=1.0, center=0.0, kt=1.0):
       = 1/(amplitude*exp((x - center)/kt))
    form = 'fermi' is the Fermi-Dirac distribution
    thermal_distribution(x, amplitude=1.0, center=0.0, kt=1.0):
       = 1/(amplitude*exp((x - center)/kt) + 1)

    Notes:
    - ``kt`` should be defined in the same units as ``x``. (The Boltzmann
      constant is kB = 8.617e-5 eV/K).
    - set ``kt<0`` to implement the energy loss convention common in scattering
      research.

    see http://hyperphysics.phy-astr.gsu.edu/hbase/quantum/disfcn.html

    """
    form = form.lower()
    if form.startswith('bose'):
        offset = -1
    elif form.startswith('maxwell'):
        offset = 0
    elif form.startswith('fermi'):
        offset = 1
    else:
        msg = "Invalid value ('%s') for argument 'form'; should be one of %s."\
              % (form, "'maxwell', 'fermi', or 'bose'")
        raise ValueError(msg)

    return real(1/(amplitude*exp((x - center)/kt) + offset + tiny*1j))


def step(x, amplitude=1.0, center=0.0, sigma=1.0, form='linear'):
    """Return a step function.

    starts at 0.0, ends at amplitude, with half-max at center, and
    rising with form:
      'linear' (default) = amplitude * min(1, max(0, arg))
      'atan', 'arctan'   = amplitude * (0.5 + atan(arg)/pi)
      'erf'              = amplitude * (1 + erf(arg))/2.0
      'logistic'         = amplitude * [1 - 1/(1 + exp(arg))]

    where arg = (x - center)/sigma

    """
    out = (x - center)/max(tiny, sigma)

    if form == 'erf':
        out = 0.5*(1 + erf(out))
    elif form == 'logistic':
        out = (1. - 1./(1. + exp(out)))
    elif form in ('atan', 'arctan'):
        out = 0.5 + arctan(out)/pi
    elif form == 'linear':
        out[where(out < 0)] = 0.0
        out[where(out > 1)] = 1.0
    else:
        msg = "Invalid value ('%s') for argument 'form'; should be one of %s."\
               % (form, "'erf', 'logistic', 'atan', 'arctan', or 'linear'")
        raise ValueError(msg)

    return amplitude*out


def rectangle(x, amplitude=1.0, center1=0.0, sigma1=1.0,
              center2=1.0, sigma2=1.0, form='linear'):
    """Return a rectangle function: step up, step down.

    (see step function)
    starts at 0.0, rises to amplitude (at center1 with width sigma1)
    then drops to 0.0 (at center2 with width sigma2) with form:
      'linear' (default) = ramp_up + ramp_down
      'atan', 'arctan'   = amplitude*(atan(arg1) + atan(arg2))/pi
      'erf'              = amplitude*(erf(arg1) + erf(arg2))/2.
      'logisitic'        = amplitude*[1 - 1/(1 + exp(arg1)) - 1/(1+exp(arg2))]

    where arg1 =  (x - center1)/sigma1
    and   arg2 = -(x - center2)/sigma2

    """
    arg1 = (x - center1)/max(tiny, sigma1)
    arg2 = (center2 - x)/max(tiny, sigma2)

    if form == 'erf':
        out = 0.5*(erf(arg1) + erf(arg2))
    elif form == 'logistic':
        out = (1. - 1./(1. + exp(arg1)) - 1./(1. + exp(arg2)))
    elif form in ('atan', 'arctan'):
        out = (arctan(arg1) + arctan(arg2))/pi
    elif form == 'linear':
        arg1[where(arg1 < 0)] = 0.0
        arg1[where(arg1 > 1)] = 1.0
        arg2[where(arg2 > 0)] = 0.0
        arg2[where(arg2 < -1)] = -1.0
        out = arg1 + arg2
    else:
        msg = "Invalid value ('%s') for argument 'form'; should be one of %s."\
               % (form, "'erf', 'logistic', 'atan', 'arctan', or 'linear'")
        raise ValueError(msg)

    return amplitude*out


def exponential(x, amplitude=1, decay=1):
    """Return an exponential function.

    x -> amplitude * exp(-x/decay)

    """
    decay = max(tiny, decay)
    return amplitude * exp(-x/decay)


def powerlaw(x, amplitude=1, exponent=1.0):
    """Return the powerlaw function.

    x -> amplitude * x**exponent

    """
    return amplitude * x**exponent


def linear(x, slope=1.0, intercept=0.0):
    """Return a linear function.

    x -> slope * x + intercept

    """
    return slope * x + intercept


def parabolic(x, a=0.0, b=0.0, c=0.0):
    """Return a parabolic function.

    x -> a * x**2 + b * x + c

    """
    return a * x**2 + b * x + c


def polynd(x, *args):
    return polyval([*args], x)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul 29 18:40:31 2020

@author: xiao
"""

SKLEARN_OPT = {"NEIGHBOR":{"KDE":{'algorithm':['auto', 'kd_tree', 'ball_tree'],
                                  'kernel':['gaussian', 'tophat', 'epanechnikov', 'exponential', 'linear', 'cosine'],
                                  'metric':['euclidean'],
                                  'breadth_first':True,
                                  'bandwidth':1.0,
                                  'atol':0,
                                  'rtol':1e-8,
                                  'leaf_size':40}},
               "CLUSTER":{},
               "REGRES":{},
               "CLASSIF":{},
               "DECOMP":{},
               "PREPROC":{}}


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Aug 24 08:56:30 2020

@author: xiao
"""

import os, functools, inspect

from pathlib import Path
import h5py, tifffile
import numpy as np

from ..dicts.config_dict import (IO_TOMO_CFG_DEFAULT,
                                 IO_XANES2D_CFG_DEFAULT,
                                 IO_XANES3D_CFG_DEFAULT)


def sli_interpreter(sli_tuple):
    sli = []
    for s in sli_tuple:
        if isinstance(s, list):
            sli.append(slice(s[0], s[1]))
        else:
            sli.append(slice(s))
    return tuple(sli)


def tomo_h5_reader(fn, dtype='data', sli=[[None, None]], cfg=IO_TOMO_CFG_DEFAULT):
    """
    Inputs:
        fnt: string
             file name template to raw h5 data files
        cfg: dictionary; optional
             dictionary with 4 items 'data_path', 'flat_path', 'dark_path', and
             'eng_path' under item 'io_data_structure'; paths point to the
             datasets locations in h5 files
        dtype: string; optional
             specifying which one of the above four items will be read
        sli: list; optional
             a 4-element list; each element is again a list indicating the
             start and end of the slicing on that dimension. The four dimensions
             are [eng, angle, y, x] for dtype='data', or three dimensions of
             [eng, y, x] for dtype='flat' and 'dark', or one dimension of [eng]
             for dtype='eng'
    Returns:
        ndarray
    """
    sli = sli_interpreter(sli)
    with h5py.File(fn, 'r') as f:
        if dtype == 'data':
            # print(sli)
            return np.squeeze(f[cfg['structured_h5_reader']['io_data_structure']['data_path']])[sli]
        elif dtype == 'flat':
            return np.squeeze(f[cfg['structured_h5_reader']['io_data_structure']['flat_path']])[sli]
        elif dtype == 'dark':
            return np.squeeze(f[cfg['structured_h5_reader']['io_data_structure']['dark_path']])[sli]
        elif dtype == 'theta':
            return np.squeeze(f[cfg['structured_h5_reader']['io_data_structure']['theta_path']])[sli[0]]


def tomo_h5_info(fn, dtype='data', cfg=IO_XANES2D_CFG_DEFAULT):
    with h5py.File(fn,"r") as f:
        if dtype == 'data':
            try:
                arr = np.squeeze(f[cfg['structured_h5_reader']['io_data_structure']['data_path']])
                dim = arr.shape
                return dim
            except:
                return 0
        elif dtype == 'flat':
            try:
                arr = np.squeeze(f[cfg['structured_h5_reader']['io_data_structure']['flat_path']])
                dim = arr.shape
                return dim
            except:
                return 0
        elif dtype == 'dark':
            try:
                arr = np.squeeze(f[cfg['structured_h5_reader']['io_data_structure']['dark_path']])
                dim = arr.shape
                return dim
            except:
                return 0
        elif dtype == 'theta':
            try:
                arr = np.squeeze(f[cfg['structured_h5_reader']['io_data_structure']['theta_path']])
                dim = arr.shape
                return dim
            except:
                return 0


def xanes2D_h5_reader(fn, dtype='data', sli=[[None, None]], cfg=IO_XANES2D_CFG_DEFAULT):
    sli = sli_interpreter(sli)
    with h5py.File(fn, 'r') as f:
        if dtype == 'data':
            return np.squeeze(f[cfg['structured_h5_reader']['io_data_structure']['data_path']])[sli]
        elif dtype == 'flat':
            return np.squeeze(f[cfg['structured_h5_reader']['io_data_structure']['flat_path']])[sli]
        elif dtype == 'dark':
            return np.squeeze(f[cfg['structured_h5_reader']['io_data_structure']['dark_path']])[sli]
        elif dtype == 'eng':
            return np.squeeze(f[cfg['structured_h5_reader']['io_data_structure']['eng_path']])[sli]


def xanes3D_h5_reader(fn, ids, dtype='data', sli=[[None, None]], cfg=IO_XANES3D_CFG_DEFAULT):
    """
    Inputs:
        fn:  string
             file name template to raw h5 data files
        cfg: dictionary; optional
             dictionary with 4 items 'data_path', 'flat_path', 'dark_path', and
             'eng_path' under item 'io_data_structure'; paths point to the
             datasets locations in h5 files
        dtype: string; optional
             specifying which one of the above four items will be read
        sli: list; optional
             a 4-element list; each element is again a list indicating the
             start and end of the slicing on that dimension. The four dimensions
             are [eng, angle, y, x] for dtype='data', or three dimensions of
             [eng, y, x] for dtype='flat' and 'dark', or one dimension of [eng]
             for dtype='eng'
    Returns:
        ndarray
    """
    sli = sli_interpreter(sli)
    if dtype == 'data':
        data = []
        for ii in ids:
            fn0 = fn.format(ii)
            with h5py.File(fn0, 'r') as f:
                data.append(np.squeeze(f[cfg['structured_h5_reader']
                              ['io_data_structure']['data_path']])[sli])
        return np.array(data)
    elif dtype == 'flat':
        data = []
        for ii in ids:
            fn0 = fn.format(ii)
            with h5py.File(fn0, 'r') as f:
                data.append(np.squeeze(f[cfg['structured_h5_reader']
                              ['io_data_structure']['flat_path']])[sli])
        return np.array(data)
    elif dtype == 'dark':
        data = []
        for ii in ids:
            fn0 = fn.format(ii)
            with h5py.File(fn0, 'r') as f:
                data.append(np.squeeze(f[cfg['structured_h5_reader']
                              ['io_data_structure']['dark_path']])[sli])
        return np.array(data)
    elif dtype == 'eng':
        eng = []
        for ii in ids:
            fn0 = fn.format(ii)
            with h5py.File(fn0, 'r') as f:
                eng.append(np.squeeze(f[cfg['structured_h5_reader']
                             ['io_data_structure']['eng_path']])[()])
        return np.array(eng)


def data_reader(func):
    @functools.wraps(func)
    def wrapper_data_reader(fn, *args, **kwargs):
        return func(fn, *args, **kwargs)
    return wrapper_data_reader


def data_info(func):
    @functools.wraps(func)
    def wrapper_data_info(fn, dtype='data', *args, **kwargs):
        if not (set(('fn', 'dtype')) <= set(inspect.getfullargspec(func).args)):
            raise TypeError(f'{func.__name__} missing one or more arguments in ("fn", "dtype")')
            return None
        return func(fn, dtype=dtype, *args, **kwargs)
    return wrapper_data_info


def read_xanes2D_wl_fit(fn):
    with h5py.File(fn, 'r') as f:
        return f['/processed_XANES2D/proc_spectrum/whiteline_pos_fit'][:]


def tiff_vol_reader(fn_temp, scan_id, roi):
    vol = []
    for ii in range(roi[0], roi[1]):
        vol.append(tifffile.imread(fn_temp.format(scan_id, str(ii).zfill(5)))[roi[2]:roi[3], roi[4]:roi[5]])
    return np.array(vol)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jul 19 22:58:16 2020

@author: xiao
"""

from ipywidgets import widgets
from copy import deepcopy
import os, h5py, json
import numpy as np

import napari
napari.gui_qt()

from .gui_components import (SelectFilesButton, NumpyArrayEncoder, get_handles, enable_disable_boxes, 
                             gen_external_py_script, fiji_viewer_off, scale_eng_list)
from ..dicts.xanes_analysis_dict import XANES_ANA_METHOD

class xanes_analysis_gui():
    def __init__(self, parent_h, form_sz=[650, 740]):
        self.parent_h = parent_h
        self.hs = {}
        self.form_sz = form_sz    
        
        self.ana_fn_seled = False
        self.ana_data_existed = False
        self.ana_data_configed = False
        
        self.ana_proc_path_in_h5 = None
        self.ana_proc_data_list = [""]
        self.ana_data = None
        
    def lock_message_text_boxes(self):
        boxes = ['AnaDataSetupInfo text']
        enable_disable_boxes(self.hs, boxes, disabled=True, level=-1)
        
    def boxes_logic(self):
        if (not self.ana_fn_seled):
            boxes = ['AnaDataSetupItem list',
                     'AnaDataSetupRead btn',
                     'AnaConfig box',
                     'AnaDataDisp box']
            enable_disable_boxes(self.hs, boxes, disabled=True, level=-1)
        elif (self.ana_fn_seled & (not self.ana_data_existed)):
            boxes = ['AnaDataSetupItem list',
                     'AnaDataSetupRead btn',
                     'AnaConfig box',
                     'AnaDataDisp box']
            enable_disable_boxes(self.hs, boxes, disabled=True, level=-1)
        elif ((self.ana_fn_seled & self.ana_data_existed) & 
              (not self.ana_data_configed)):
            boxes = ['AnaConfig box',
                     'AnaDataDisp box']
            enable_disable_boxes(self.hs, boxes, disabled=True, level=-1)
            boxes = ['AnaDataSetupItem list',
                     'AnaDataSetupRead btn']
            enable_disable_boxes(self.hs, boxes, disabled=False, level=-1)
        elif ((self.ana_fn_seled & self.ana_data_existed) & 
              self.ana_data_configed):
            boxes = ['AnaDataSetupItem list',
                     'AnaDataSetupRead btn',
                     'AnaConfig box',
                     'AnaDataDisp box']
            enable_disable_boxes(self.hs, boxes, disabled=False, level=-1)
        
    def build_gui(self):
        self.hs['Ana form'] = widgets.Tab()
        self.hs['AnaData tab'] = widgets.VBox()
        self.hs['AnaPrep tab'] = widgets.VBox()
        self.hs['AnaConfig tab'] = widgets.VBox()
        self.hs['AnaDisp tab'] = widgets.VBox()
        self.hs['Ana form'].children = [self.hs['AnaData tab'],
                                        self.hs['AnaPrep tab'],
                                        self.hs['AnaConfig tab'],
                                        self.hs['AnaDisp tab']]
        layout = {'border':'3px solid #FFCC00', 'width':f'{0.96*self.form_sz[1]-98}px', 
                  'height':f'{0.89*(self.form_sz[0]-128)}px'}
        self.hs['AnaData tab'].layout = layout
        self.hs['AnaConfig tab'].layout = layout
        self.hs['AnaDisp tab'].layout = layout
        self.hs['Ana form'].set_title(0, 'Data Setup')
        self.hs['Ana form'].set_title(1, 'Data PreProc')
        self.hs['Ana form'].set_title(2, 'Ana Config')
        self.hs['Ana form'].set_title(3, 'Ana Disp')
        
        ## ## ## ## ## data setup -- start
        self.hs['DataSetup box'] = widgets.HBox()
        
        ## ## ## ## ## ## config parameters -- start
        data_setup_GridSpecLayout = widgets.GridspecLayout(16, 200,
                                                           layout={"border":"3px solid #FFCC00",
                                                                   "width":f"{0.96*self.form_sz[1]-98}px",
                                                                   "height":f"{0.88*(self.form_sz[0]-128)}px",
                                                                   "align_items":"flex-start",
                                                                   "justify_items":"flex-start"})           
        data_setup_GridSpecLayout[4, 5:25] = SelectFilesButton(option='askopenfilename',
                                                               description_tip="Open h5 file in which the processed XANES data is available",
                                                               **{'open_filetypes': (('h5 files', '*.h5'),)},
                                                               initialdir=self.parent_h.global_h.cwd)
        self.hs['AnaDataSetupOpen btn'] = data_setup_GridSpecLayout[4, 5:25]
        self.hs['AnaDataSetupOpen btn'].style.button_color = 'darkviolet'
        self.hs['AnaDataSetupOpen btn'].disabled = True
        data_setup_GridSpecLayout[:8, 30:150] = widgets.Select(options=[''],
                                                               value='',
                                                               rows=8,
                                                               description='item:',
                                                               disabled=True,
                                                               layout={'width':'100%', 'height':'100%'})
        self.hs['AnaDataSetupItem list'] = data_setup_GridSpecLayout[:8, 30:150]
        data_setup_GridSpecLayout[4, 160:180] = widgets.Button(description="Read Item",
                                                               description_tip="Read in the selected item",
                                                               disabled = True) 
        self.hs['AnaDataSetupRead btn'] = data_setup_GridSpecLayout[4, 160:180]
        self.hs['AnaDataSetupRead btn'].style.button_color = 'darkviolet'
        data_setup_GridSpecLayout[9:, :] = widgets.Textarea(value='Data Info',
                                                            placeholder='Data Info',
                                                            description='Data Info',
                                                            disabled=True,
                                                            layout={'width':'90%', 'height':'95%'})
        self.hs['AnaDataSetupInfo text'] = data_setup_GridSpecLayout[9:, :]
        
        self.hs['AnaDataSetupOpen btn'].on_click(self.ana_data_setup_open_btn_clk)
        self.hs['AnaDataSetupItem list'].observe(self.ana_data_setup_item_list_chg, names='value')
        self.hs['AnaDataSetupRead btn'].on_click(self.ana_data_setup_read_btn_clk)

        self.hs['DataSetup box'] = data_setup_GridSpecLayout
        self.hs['DataSetup box'].children = [self.hs['AnaDataSetupOpen btn'],
                                             self.hs['AnaDataSetupItem list'],
                                             self.hs['AnaDataSetupOpen btn'],
                                             self.hs['AnaDataSetupInfo text']]
        ## ## ## ## ## ## config parameters -- end
        
        self.hs['AnaData tab'].children = [self.hs['DataSetup box']]
        ## ## ## ## ## data setup -- end
        
        ## ## ## ## ##  config analysis -- start
        self.hs['AnaConfig box'] = widgets.HBox()
        
        ## ## ## ## ## ## config parameters -- start
        ana_setup_GridSpecLayout = widgets.GridspecLayout(16, 200,
                                                          layout = {"border":"3px solid #FFCC00",
                                                                    'width':f'{0.96*self.form_sz[1]-98}px',
                                                                    "height":f"{0.88*(self.form_sz[0]-128)}px",
                                                                    "align_items":"flex-start",
                                                                    "justify_items":"flex-start"})
        ana_setup_GridSpecLayout[0, 0:50] = widgets.Dropdown(description='ana type',
                                                             description_tooltip='choose analysis to apply to the selected data',
                                                             options = ['Preproc', 
                                                                        'Decomp', 
                                                                        'Classif',
                                                                        'Regres',
                                                                        'Cluster',
                                                                        'Neighbor'],
                                                             value ='Neighbor',
                                                             disabled=True,
                                                             layout={'width':f'{0.3*self.form_sz[1]-98}px'})
        self.hs['AnaType drpdn'] = ana_setup_GridSpecLayout[0, 0:50]
        ana_setup_GridSpecLayout[0, 50:100] = widgets.Dropdown(description='ana method',
                                                                            description_tooltip='choose analysis to apply to the selected data',
                                                                            options = ['KDE', 'PCA', 'KNN'],
                                                                            value ='KDE',
                                                                            disabled=True,
                                                                            layout={'width':f'{0.3*self.form_sz[1]-98}px'})
        self.hs['AnaMethod drpdn'] = ana_setup_GridSpecLayout[0, 50:100]
        
        for ii in range(2):
            for jj in range(4):
                ana_setup_GridSpecLayout[ii+1, jj*50:(jj+1)*50] = \
                    widgets.Dropdown(description = 'p{0}'.format(str(ii*4+jj).zfill(2)),
                                     value = 'linear',
                                     options = ['linear'],
                                     description_tooltip = "analysis function variable 0",
                                     layout = {'width':f'{0.3*self.form_sz[1]-98}px'},
                                     disabled=True)
                self.hs['AnaConfigPars{1}'.format(1+ii*4+jj, ii*4+jj)] = \
                    ana_setup_GridSpecLayout[ii+1, jj*50:(jj+1)*50]
        
        for ii in range(2, 4):
            for jj in range(4):
                ana_setup_GridSpecLayout[ii+1, jj*50:(jj+1)*50] = \
                    widgets.BoundedFloatText(description = 'p{0}'.format(str(ii*4+jj).zfill(2)),
                                             value = 0,
                                             min = -1e5,
                                             max = 1e5,
                                             description_tooltip = "analysis function variable 0",
                                             layout = {'width':f'{0.3*self.form_sz[1]-98}px'},
                                             disabled=True)
                self.hs['AnaConfigPars{1}'.format(1+ii*4+jj, ii*4+jj)] = \
                    ana_setup_GridSpecLayout[ii+1, jj*50:(jj+1)*50]
        
        ana_setup_GridSpecLayout[6, 10:190] = \
            widgets.IntProgress(value=0,
                                min=0,
                                max=100,
                                step=1,
                                description='Completing:',
                                bar_style='info', # 'success', 'info', 'warning', 'danger' or ''
                                orientation='horizontal', 
                                indent=False,
                                layout={'width':'100%', 'height':'90%'})
        self.hs['AnaConfigPrgr bar'] = \
            ana_setup_GridSpecLayout[6, 10:190]
            
        ana_setup_GridSpecLayout[7, 90:110] = \
            widgets.Button(description='Compute',
                           description_tip='Perform the analysis',
                           disabled=True)
        self.hs['AnaConfigCmpt btn'] = \
            ana_setup_GridSpecLayout[7, 90:110]
        self.hs['AnaConfigCmpt btn'].style.button_color = 'darkviolet'
        
        self.hs['AnaType drpdn'].observe(self.ana_type_drpdn_chg, names='value')
        self.hs['AnaMethod drpdn'].observe(self.ana_meth_drpdn_chg, names='value')
        self.hs['AnaConfigPars0'].observe(self.ana_config_p0_chg, names='value')
        self.hs['AnaConfigPars1'].observe(self.ana_config_p1_chg, names='value')
        self.hs['AnaConfigPars2'].observe(self.ana_config_p2_chg, names='value')
        self.hs['AnaConfigPars3'].observe(self.ana_config_p3_chg, names='value')
        self.hs['AnaConfigPars4'].observe(self.ana_config_p4_chg, names='value')
        self.hs['AnaConfigPars5'].observe(self.ana_config_p5_chg, names='value')
        self.hs['AnaConfigPars6'].observe(self.ana_config_p6_chg, names='value')
        self.hs['AnaConfigPars7'].observe(self.ana_config_p7_chg, names='value')
        self.hs['AnaConfigCmpt btn'].on_click(self.ana_config_cmpt_btn_clk)
        self.hs['AnaConfig box'] = ana_setup_GridSpecLayout
        self.hs['AnaConfig box'].children = [self.hs['AnaType drpdn'],
                                             self.hs['AnaMethod drpdn'],
                                             self.hs['AnaConfigPars0'],
                                             self.hs['AnaConfigPars1'],
                                             self.hs['AnaConfigPars2'],
                                             self.hs['AnaConfigPars3'],
                                             self.hs['AnaConfigPars4'],
                                             self.hs['AnaConfigPars5'],
                                             self.hs['AnaConfigPars6'],
                                             self.hs['AnaConfigPars7'],
                                             self.hs['AnaConfigCmpt btn']]
        ## ## ## ## ## ## config parameters -- end
        
        self.hs['AnaConfig tab'].children = [self.hs['AnaConfig box']]
        ## ## ## ## ## config analysis -- end
        
        ## ## ## ## ##  result display -- start
        self.hs['AnaDataDisp box'] = widgets.HBox()
        ## ## ## ## ##  result display -- end
        
    def ana_data_setup_open_btn_clk(self, a):
        if len(a.files[0]) != 0:
            self.ana_fn_seled = True
            self.ana_fn = os.path.abspath(a.files[0])
            with h5py.File(self.ana_fn, 'r') as f:
                for key in f['/']:
                    if 'processed_' in key:
                        self.ana_proc_path_in_h5 = key + '/proc_spectrum'
                        self.ana_data_existed = True
                        break
                    else:
                        self.ana_data_existed = False
                if self.ana_data_existed:
                    self.ana_proc_data_list = []
                    for key in f[self.ana_proc_path_in_h5]:
                        self.ana_proc_data_list.append(key)
                    self.hs['AnaDataSetupItem list'].options = \
                        self.ana_proc_data_list
                else:
                    self.ana_proc_data_list = [""]
        else:
            self.ana_fn_seled = False
            self.ana_proc_data_list = [""]
            self.hs['AnaDataSetupItem list'].options = \
                self.ana_proc_data_list
            self.hs['AnaDataSetupItem list'].value = \
                self.ana_proc_data_list[0]  
        self.boxes_logic()
        self.lock_message_text_boxes()
        
    def ana_data_setup_item_list_chg(self, a):
        with h5py.File(self.ana_fn, 'r') as f:
            try:
                shape = f[self.ana_proc_path_in_h5+'/'+
                          self.hs['AnaDataSetupItem list'].value].shape
            except:
                shape = None
            try:
                dtype = f[self.ana_proc_path_in_h5+'/'+
                          self.hs['AnaDataSetupItem list'].value].dtype
            except:
                dtype = None
            try:
                dmin = np.min(f[self.ana_proc_path_in_h5+'/'+
                                self.hs['AnaDataSetupItem list'].value])
            except:
                dmin = None
            try:
                dmax = np.max(f[self.ana_proc_path_in_h5+'/'+
                          self.hs['AnaDataSetupItem list'].value])
            except:
                dmax = None
        self.hs['AnaDataSetupInfo text'].value = 'shape: ' + str(shape) + '\n' +\
                                                                         'dtype: ' + str(dtype) + '\n' +\
                                                                         'data min: ' +str(dmin) + '\n' +\
                                                                         'data max: ' +str(dmax) 
        self.boxes_logic()
        self.lock_message_text_boxes()
    
    def ana_data_setup_read_btn_clk(self, a):
        with h5py.File(self.ana_fn, 'r') as f:
            self.ana_data = f[self.ana_proc_path_in_h5+'/'+
                              self.hs['AnaDataSetupItem list'].value][:]
            self.ana_data_configed = True
        self.boxes_logic()
        self.lock_message_text_boxes()
    
    def ana_type_drpdn_chg(self, a):
        pass
    
    def ana_meth_drpdn_chg(self, a):
        pass
    
    def ana_config_p0_chg(self, a):
        pass
    
    def ana_config_p1_chg(self, a):
        pass
    
    def ana_config_p2_chg(self, a):
        pass
    
    def ana_config_p3_chg(self, a):
        pass
    
    def ana_config_p4_chg(self, a):
        pass
    
    def ana_config_p5_chg(self, a):
        pass
    
    def ana_config_p6_chg(self, a):
        pass
    
    def ana_config_p7_chg(self, a):
        pass
    
    def ana_config_cmpt_btn_clk(self, a):
        pass
        
        
        